from unittest import TestCase
from hamcrest import assert_that, is_, instance_of
from mock import MagicMock
import os
from cis.data_io.ungridded_data import UngriddedData, UngriddedDataList
from cis.data_io.gridded_data import GriddedData, GriddedDataList, make_from_cube
from cis.data_io.data_reader import DataReader
from cis.test.util.mock import make_square_5x3_2d_cube, make_regular_2d_ungridded_data


class TestDataReader(TestCase):
    def test_GIVEN_single_variable_gridded_WHEN_read_data_THEN_GriddedData_returned(self):
        variables = 'var1'
        filenames = 'filename1'
        product = None
        gridded_data = make_square_5x3_2d_cube()
        gridded_data.__class__ = GriddedData
        get_data_func = MagicMock(return_value=gridded_data)
        reader = DataReader(get_data_func=get_data_func)
        data = reader.read_data_list(filenames, variables, product)[0]

        # Check the data read function is called correctly
        assert_that(get_data_func.call_count, is_(1))
        call_args = get_data_func.call_args_list[0][0]
        assert_that(call_args[0], is_([filenames]))
        assert_that(call_args[1], is_(variables))
        assert_that(call_args[2], is_(product))

        # Check the data we got back is as expected
        assert_that(data, instance_of(GriddedData))
        assert_that(data.data.tolist(), is_(make_square_5x3_2d_cube().data.tolist()))

    def test_GIVEN_single_variable_ungridded_WHEN_read_data_THEN_GriddedData_returned(self):
        variables = 'var1'
        filenames = 'filename1'
        product = None
        get_data_func = MagicMock(return_value=make_regular_2d_ungridded_data())
        reader = DataReader(get_data_func=get_data_func)
        data = reader.read_data_list(filenames, variables, product)[0]

        # Check the data read function is called correctly
        assert_that(get_data_func.call_count, is_(1))
        call_args = get_data_func.call_args_list[0][0]
        assert_that(call_args[0], is_([filenames]))
        assert_that(call_args[1], is_(variables))
        assert_that(call_args[2], is_(product))

        # Check the data we got back is as expected
        assert_that(data, instance_of(UngriddedData))
        assert_that(data.data.tolist(), is_(make_regular_2d_ungridded_data().data.tolist()))

    def test_GIVEN_multiple_variable_gridded_WHEN_read_data_THEN_GriddedDataList_returned(self):
        variables = ['var1', 'var2']
        filenames = 'filename1'
        product = None
        gridded_data = make_square_5x3_2d_cube()
        gridded_data.__class__ = GriddedData
        get_data_func = MagicMock(return_value=gridded_data)
        reader = DataReader(get_data_func=get_data_func)
        data = reader.read_data_list(filenames, variables, product)

        # Check the data read function is called correctly
        assert_that(get_data_func.call_count, is_(2))
        first_call_args = get_data_func.call_args_list[0][0]
        second_call_args = get_data_func.call_args_list[1][0]
        assert_that(first_call_args[0], is_([filenames]))
        assert_that(first_call_args[1], is_(variables[0]))
        assert_that(second_call_args[1], is_(variables[1]))
        assert_that(first_call_args[2], is_(product))

        # Check the data we got back is as expected
        assert_that(data, instance_of(GriddedDataList))
        assert_that(data[0].data.tolist(), is_(make_square_5x3_2d_cube().data.tolist()))
        assert_that(data[1].data.tolist(), is_(data[0].data.tolist()))

    def test_GIVEN_multiple_variable_ungridded_WHEN_read_data_THEN_GriddedDataList_returned(self):
        variables = ['var1', 'var2']
        filenames = 'filename1'
        product = None
        get_data_func = MagicMock(return_value=make_regular_2d_ungridded_data())
        reader = DataReader(get_data_func=get_data_func)
        data = reader.read_data_list(filenames, variables, product)

        # Check the data read function is called correctly
        assert_that(get_data_func.call_count, is_(2))
        first_call_args = get_data_func.call_args_list[0][0]
        second_call_args = get_data_func.call_args_list[1][0]
        assert_that(first_call_args[0], is_([filenames]))
        assert_that(first_call_args[1], is_(variables[0]))
        assert_that(second_call_args[1], is_(variables[1]))
        assert_that(first_call_args[2], is_(product))

        # Check the data we got back is as expected
        assert_that(data, instance_of(UngriddedDataList))
        assert_that(data[0].data.tolist(), is_(make_regular_2d_ungridded_data().data.tolist()))
        assert_that(data[1].data.tolist(), is_(data[0].data.tolist()))

    def test_GIVEN_multiple_variable_mix_of_gridded_ungridded_WHEN_read_data_THEN_raises_TypeError(self):
        variables = ['var1', 'var2']
        filenames = 'filename1'
        product = None
        gridded_data = make_square_5x3_2d_cube()
        gridded_data.__class__ = GriddedData
        ungridded_data = make_regular_2d_ungridded_data()
        get_data_func = MagicMock(side_effect=[gridded_data, ungridded_data])
        reader = DataReader(get_data_func=get_data_func)
        with self.assertRaises(TypeError):
            data = reader.read_data_list(filenames, variables, product)[0]

    def test_GIVEN_wildcards_WHEN_read_data_THEN_matching_variables_identified(self):
        variables = ['*.nc', 'test?.hdf']
        file_vars = ['aeronet.lev20', 'var2.hdf', 'netcdf1.nc', 'netcdf3.nc', 'test.hdf', 'test1.hdf']
        should_match = ['netcdf1.nc', 'netcdf3.nc', 'test1.hdf']
        filenames = 'filename1'
        get_data_func = MagicMock(return_value=make_regular_2d_ungridded_data())
        get_var_func = MagicMock(return_value=file_vars)
        reader = DataReader(get_data_func=get_data_func, get_variables_func=get_var_func)
        reader.read_data_list(filenames, variables)[0]
        assert_that(reader._get_data_func.call_count, is_(len(should_match)))
        for i in range(len(should_match)):
            assert_that(reader._get_data_func.call_args_list[i][0][1], is_(should_match[i]))

    def test_GIVEN_no_matching_variables_for_wildcards_WHEN_read_data_THEN_no_Error(self):
        variables = ['aeronet.lev20', '*.nc', 'test?.hdf']
        file_vars = ['aeronet.lev20', 'var2.hdf']
        filenames = 'filename1'
        get_data_func = MagicMock(return_value=make_regular_2d_ungridded_data())
        get_var_func = MagicMock(return_value=file_vars)
        reader = DataReader(get_data_func=get_data_func, get_variables_func=get_var_func)
        reader.read_data_list(filenames, variables)[0]
        assert_that(reader._get_data_func.call_count, is_(1))
        assert_that(reader._get_data_func.call_args_list[0][0][1], is_('aeronet.lev20'))

    def test_GIVEN_no_matching_variables_found_overall_WHEN_read_data_THEN_raises_Error(self):
        variables = ['test?.hdf', '*.nc']
        file_vars = ['sample_file.hdf', 'aeronet.lev20']
        filenames = 'filename1'
        get_data_func = MagicMock()
        get_var_func = MagicMock(return_value=file_vars)
        reader = DataReader(get_data_func=get_data_func, get_variables_func=get_var_func)
        with self.assertRaises(ValueError):
            reader.read_data_list(filenames, variables)[0]

    def test_GIVEN_multiple_datagroups_WHEN_read_datagroups_THEN_get_data_called_correctly(self):
        datagroup_1 = {'variables': ['var1', 'var2'],
                       'filenames': ['filename1.nc'],
                       'product': None}
        datagroup_2 = {'variables': ['var3', 'var4'],
                       'filenames': ['filename2.nc'],
                       'product': 'cis'}
        get_data_func = MagicMock(return_value=make_regular_2d_ungridded_data())
        get_var_func = MagicMock(side_effect=lambda f: {'filename1.nc': ['var1', 'var2'],
                                                        'filename2.nc': ['var3', 'var4']}[f])
        reader = DataReader(get_data_func=get_data_func, get_variables_func=get_var_func)
        data = reader.read_datagroups([datagroup_1, datagroup_2])
        assert_that(get_data_func.call_count, is_(4))
        assert_that(get_data_func.call_args_list[0][0], is_((['filename1.nc'], 'var1', None)))
        assert_that(get_data_func.call_args_list[1][0], is_((['filename1.nc'], 'var2', None)))
        assert_that(get_data_func.call_args_list[2][0], is_((['filename2.nc'], 'var3', 'cis')))
        assert_that(get_data_func.call_args_list[3][0], is_((['filename2.nc'], 'var4', 'cis')))

    def test_GIVEN_ungridded_datagroups_with_different_num_vars_WHEN_read_datagroups_THEN_data_returned_in_list(self):
        datagroup_1 = {'variables': ['var1', 'var2'],
                       'filenames': ['filename1.nc'],
                       'product': None}
        datagroup_2 = {'variables': ['var3'],
                       'filenames': ['filename2.nc'],
                       'product': 'cis'}
        var1 = make_regular_2d_ungridded_data()
        var2 = make_regular_2d_ungridded_data()
        var3 = make_regular_2d_ungridded_data()
        get_data_func = MagicMock(side_effect=[var1, var2, var3])
        get_var_func = MagicMock(side_effect=lambda f: {'filename1.nc': ['var1', 'var2'],
                                                        'filename2.nc': ['var3']}[f])
        reader = DataReader(get_data_func=get_data_func, get_variables_func=get_var_func)
        data = reader.read_datagroups([datagroup_1, datagroup_2])
        assert_that(len(data), is_(3))
        assert_that(data[0], is_(var1))
        assert_that(data[1], is_(var2))
        assert_that(data[2], is_(var3))

    def test_GIVEN_gridded_datagroups_WHEN_read_datagroups_THEN_data_returned_in_list(self):
        datagroup_1 = {'variables': ['var1', 'var2'],
                       'filenames': ['filename1.nc'],
                       'product': None}
        datagroup_2 = {'variables': ['var3'],
                       'filenames': ['filename2.nc'],
                       'product': 'cis'}
        var1 = make_from_cube(make_square_5x3_2d_cube())
        var2 = make_from_cube(make_square_5x3_2d_cube())
        var3 = make_from_cube(make_square_5x3_2d_cube())
        get_data_func = MagicMock(side_effect=[var1, var2, var3])
        get_var_func = MagicMock(side_effect=lambda f: {'filename1.nc': ['var1', 'var2'],
                                                        'filename2.nc': ['var3']}[f])
        reader = DataReader(get_data_func=get_data_func, get_variables_func=get_var_func)
        data = reader.read_datagroups([datagroup_1, datagroup_2])
        assert_that(len(data), is_(3))
        assert_that(data[0], is_(var1))
        assert_that(data[1], is_(var2))
        assert_that(data[2], is_(var3))

    def test_GIVEN_aliases_None_WHEN_read_datagroups_THEN_read_OK_aliases_default_to_var_names(self):
        datagroup = {'variables': ['var1'],
                     'filenames': ['filename1.nc'],
                     'product': None,
                     'aliases': None}
        var1 = make_from_cube(make_square_5x3_2d_cube())
        get_data_func = MagicMock(side_effect=[var1])
        get_var_func = MagicMock(side_effect=['var1'])
        reader = DataReader(get_data_func=get_data_func, get_variables_func=get_var_func)
        data = reader.read_datagroups([datagroup])
        assert_that(data[0].var_name, is_('dummy'))

    def test_GIVEN_aliases_missing_WHEN_read_datagroups_THEN_read_OK_aliases_default_to_var_names(self):
        datagroup = {'variables': ['var1'],
                     'filenames': ['filename1.nc'],
                     'product': None}
        var1 = make_from_cube(make_square_5x3_2d_cube())
        get_data_func = MagicMock(side_effect=[var1])
        get_var_func = MagicMock(side_effect=['var1'])
        reader = DataReader(get_data_func=get_data_func, get_variables_func=get_var_func)
        data = reader.read_datagroups([datagroup])
        assert_that(data[0].var_name, is_('dummy'))

    def test_GIVEN_not_enough_aliases_WHEN_read_datagroups_THEN_raises_ValueError(self):
        datagroup = {'variables': ['var1', 'var2'],
                     'filenames': ['filename1.nc'],
                     'product': None,
                     'aliases': ['alias1']}
        var1 = make_from_cube(make_square_5x3_2d_cube())
        var2 = make_from_cube(make_square_5x3_2d_cube())
        get_data_func = MagicMock(side_effect=[var1, var2])
        get_var_func = MagicMock(side_effect=['var1', 'var2'])
        reader = DataReader(get_data_func=get_data_func, get_variables_func=get_var_func)
        with self.assertRaises(ValueError):
            data = reader.read_datagroups([datagroup])

    def test_GIVEN_aliases_WHEN_read_datagroups_THEN_output_data_has_aliases(self):
        datagroup = {'variables': ['var1'],
                     'filenames': ['filename1.nc'],
                     'product': None,
                     'aliases': ['alias1']}
        get_data_func = MagicMock(return_value=make_from_cube(make_square_5x3_2d_cube()))
        get_var_func = MagicMock(return_value=['var1'])
        reader = DataReader(get_data_func=get_data_func, get_variables_func=get_var_func)
        data = reader.read_datagroups([datagroup])
        assert_that(data[0].alias, is_('alias1'))



"""
Module containing hdf file utility functions for the VD object
"""
import numpy as np
# Optional HDF import, if the module isn't found we defer raising ImportError until it is actually needed.
try:
    from pyhdf.HDF import HDF, HDF4Error
    from pyhdf.VS import VS
except ImportError:
    HDF = None

from collections import namedtuple
import logging
from cis.utils import create_masked_array_for_missing_values, listify


class VDS(namedtuple('VDS', ['filename', 'variable'])):
    pass


def get_hdf_VD_file_variables(filename):
    """
    Get all the variables from an HDF VD file

    :param filename: The filename of the file to get the variables from
    :return: An OrderedDict containing the variables from the file
    """
    variables = None
    if not HDF:
        raise ImportError("HDF support was not installed, please reinstall with pyhdf to read HDF files.")

    try:
        # Open file
        datafile = HDF(filename)
        vs = datafile.vstart()
        # List of required variable names
        names = vs.vdatainfo()
        # This returns a list of tuples, so convert into a dictionary for easy lookup
        variables = {}
        for var in names:
            variables[var[0]] = var[1:]
        # Close file
        vs.end()
        datafile.close()
    except:
        logging.error("Error while reading VD data")

    return variables


def read(filename, variables=None, datadict=None):
    """
    Given a filename and a list of file names return a dictionary of VD data handles

    :param filename: full path to a single HDF4 file
    :param variables: A list of variables to read, if no variables are given, no variables are read
    :param datadict: A dictionary of variable name, data handle pairs to be appended to
    :return: An updated datadict with any new variables appended.
    """

    if not HDF:
        raise ImportError("HDF support was not installed, please reinstall with pyhdf to read HDF files.")

    if datadict is None:
        datadict = {}

    variables = listify(variables)

    vs = None
    datafile = None
    try:
        datafile = HDF(filename)
        vs = datafile.vstart()

        for variable in variables:
            try:
                vd = vs.attach(variable)
                vd.detach()
                datadict[variable] = VDS(filename, variable)
            except:
                # ignore variable that failed
                pass
    finally:
        if vs is not None:
            vs.end()
        if datafile is not None:
            datafile.close()

    return datadict


def get_data(vds, first_record=False, missing_values=None):
    """
    Actually read the data from the VDS handle. We shouldn't need to check for HDF being installed here because the
    VDS object which is being passed to us can only have come from pyhdf.

    :param vds:
    :param first_record:
    :param missing_values:
    :return:
    """

    # get file and variable reference from tuple
    filename = vds.filename
    variable = vds.variable

    try:
        datafile = HDF(filename)
    except HDF4Error as e:
        raise IOError(e)

    vs = datafile.vstart()

    if first_record:
        vd = vs.attach(vs.next(-1))
        vd.setfields(variable)
        data = vd.read()
    else:
        # get data for that variable
        vd = vs.attach(variable)
        data = vd.read(nRec=vd.inquire()[0])

    # create numpy array from data
    data = np.array(data).flatten()

    # dealing with missing data
    if missing_values is None:
        missing_values = [_get_attribute_value(vd, 'missing')]

    data = create_masked_array_for_missing_values(data, missing_values)

    # detach and close
    vd.detach()
    vs.end()
    datafile.close()

    return data


def get_metadata(vds):
    from cis.data_io.ungridded_data import Metadata

    # get file and variable reference from tuple
    filename = vds.filename
    variable = vds.variable

    datafile = HDF(filename)
    vs = datafile.vstart()

    # get data for that variable
    vd = vs.attach(variable)

    name = variable
    misc = vd.attrinfo()

    long_name = _pop_attribute_value(misc, 'long_name', '')
    units = _pop_attribute_value(misc, 'units', '')
    factor = _pop_attribute_value(misc, 'factor')
    offset = _pop_attribute_value(misc, 'offset')
    missing = _pop_attribute_value(misc, 'missing')

    # VD data are always 1D, so the shape is simply the length of the data vector
    shape = [len(vd.read(nRec=vd.inquire()[0]))]

    # Tidy up the rest of the data in misc:
    misc = {k: v[2] for k, v in misc.items()}

    metadata = Metadata(name=name, long_name=long_name, shape=shape, units=units,
                        factor=factor, offset=offset, missing_value=missing, misc=misc)

    # detach and close
    vd.detach()
    vs.end()
    datafile.close()

    return metadata


def _get_attribute_value(vd, name, default=None):
    val = vd.attrinfo().get(name, None)
    # if the attribute is not present
    if val is None:
        return default
    else:
        # attrinfo() returns a tuple in which the value of interest is the 3rd item, hence the '[2]'
        return val[2]


def _pop_attribute_value(att_dict, name, default=None):
    val = att_dict.pop(name, None)
    # if the attribute is not present
    if val is None:
        return default
    else:
        # attrinfo() returns a tuple in which the value of interest is the 3rd item, hence the '[2]'
        return val[2]



from cis.data_io import hdf_sd as hdf_sd, hdf_vd
import cis.utils as utils
import logging


def get_hdf4_file_variables(filename, data_type=None):
    """
    Get all variables from a file containing ungridded data.
    Concatenate variable from both VD and SD data

    :param filename: The filename of the file to get the variables from
    :param data_type: String representing the HDF data type, i.e. 'VD' or 'SD'. if None, both are computed.
    """

    SD_vars = VD_vars = None

    if data_type is None or data_type.lower() == "sd":
        SD_vars = hdf_sd.get_hdf_SD_file_variables(filename)
    elif data_type is None or data_type.lower() == "vd":
        VD_vars = hdf_vd.get_hdf_VD_file_variables(filename)
    else:
        raise ValueError("Invalid data-type: %s, HDF variables must be VD or SD only" % data_type)

    return SD_vars, VD_vars


def get_hdf4_file_metadata(filename):
    """
    This returns a dictionary of file attributes, which often contains metadata information
    about the whole file. The value of each attribute can simply be a big string
    which will often need to be parsed manually thereafter.
    :param filename
    :return: dictionary of string attributes
    """
    # Optional HDF import, if the module isn't found we defer raising ImportError until it is actually needed.
    try:
        from pyhdf.SD import SD
    except ImportError:
        raise ImportError("HDF support was not installed, please reinstall with pyhdf to read HDF files.")

    return SD(filename).attributes()


def _read_hdf4(filename, variables):
    """
        A wrapper method for reading raw data from hdf4 files. This returns a dictionary of io handles
         for each VD and SD data types.

        :param filename:     A name of a file to read
        :param variables:    List of variables to read from the files

        :return: (sds_dict, vds_dict) A tuple of dictionaries, one for sds objects and another for vds
    """
    from cis.exceptions import InvalidVariableError
    from pyhdf.error import HDF4Error

    variables = utils.listify(variables)

    # I'd rather not have to make this check but for pyhdf 0.9.0 and hdf 4.2.9 on OS X the c-level read routine will at
    # some point call exit(138) when reading valid netcdf files (rather than returning a negative status).
    if not filename.endswith('.hdf'):
        raise IOError("Tried to read non HDF file: {}".format(filename))

    try:
        sds_dict = hdf_sd.read(filename, variables)

        # remove the variables identified as SD (i.e. the keys in sds_dict)
        # no need to try looking for them as VD variable
        # AND this can cause a crash in some version/implementations of the core HDF4 libraries!

        # First create a copy of the list in order for the original list to be left intact when elements are removed
        # from it, this enables the original list to be used when many files are read
        vdvariables = list(variables)
        for sds_dict_key in sds_dict:
            vdvariables.remove(sds_dict_key)

        vds_dict = hdf_vd.read(filename, vdvariables)
    except HDF4Error as e:
        raise IOError(str(e))

    for variable in variables:
        if variable not in sds_dict and variable not in vds_dict:
            raise InvalidVariableError("Could not find " + variable + " in file: " + filename)

    return sds_dict, vds_dict


def read(filenames, variables):

    sdata = {}
    vdata = {}

    for filename in filenames:

        logging.debug("reading file: " + filename)

        # reading in all variables into a 2 dictionaries:
        # sdata, key: variable name, value: list of sds
        # vdata, key: variable name, value: list of vds
        sds_dict, vds_dict = _read_hdf4(filename, variables)
        for var in list(sds_dict.keys()):
            utils.add_element_to_list_in_dict(sdata, var, sds_dict[var])
        for var in list(vds_dict.keys()):
            utils.add_element_to_list_in_dict(vdata, var, vds_dict[var])

    return sdata, vdata


def read_data(data_list, read_function):
    """
    Wrapper for calling an HDF reading function for each dataset, and then concatenating the result.

    :param list data_list: A list of data objects to read
    :param callable or str read_function: A function for reading the data, or 'SD' or 'VD' for default reading routines.
    :return: A single numpy array of concatenated data values.
    """
    if callable(read_function):
        out = utils.concatenate([read_function(i) for i in data_list])
    elif read_function == 'VD':
        out = utils.concatenate([hdf_vd.get_data(i) for i in data_list])
    elif read_function == 'SD':
        out = utils.concatenate([hdf_sd.get_data(i) for i in data_list])
    else:
        raise ValueError("Invalid read-function: {}, please supply a callable read "
                         "function, 'VD' or 'SD' only".format(read_function))
    return out


def read_metadata(data_dict, data_type):
    if data_type == 'VD':
        out = hdf_vd.get_metadata(data_dict[0])
    elif data_type == 'SD':
        out = hdf_sd.get_metadata(data_dict[0])
    else:
        raise ValueError("Invalid data-type: %s, HDF variables must be VD or SD only" % data_type)
    return out



"""
    Top level collocation objects
"""
import logging
from cis.collocation.col_implementations import moments
import six


def collocate(data, sample, collocator, constraint, kernel):
    """
    Perform the collocation.

    :param CommonData or CommonDataList data: Data to collocate
    :param CommonData sample: Sampling to collocate onto
    :param cis.collocation.col_framework.Collocator collocator: The collocator object to use
    :param cis.collocation.col_framework.Constraint constraint: The constraint object
    :param cis.collocation.col_framework.Kernel  kernel: The kernel to use
    :return CommonData: The collocated data
    :raises CoordinateNotFoundError: If the collocator was unable to compare the sample and data points
    """
    from cis.exceptions import CoordinateNotFoundError
    from time import time
    from cis import __version__

    logging.info("Collocator: " + str(collocator))
    logging.info("Kernel: " + str(kernel))

    logging.info("Collocating, this could take a while...")
    t1 = time()
    try:
        new_data = collocator.collocate(sample, data, constraint, kernel)
    except (TypeError, AttributeError) as e:
        raise CoordinateNotFoundError('Collocator was unable to compare data points, check the dimensions of each '
                                      'data set and the collocation methods chosen. \n' + str(e))

    logging.info("Completed. Total time taken: " + str(time() - t1))

    for d in new_data:
        history = "Collocated onto sampling from: " + str(getattr(sample, "filenames", "Unknown")) + " " + \
                  "\nusing CIS version " + __version__ + " " + \
                  "\nvariables: " + str(getattr(data, "var_name", "Unknown")) + " " + \
                  "\nwith files: " + str(getattr(data, "filenames", "Unknown")) + " " + \
                  "\nusing collocator: " + str(collocator) + " " + \
                  "\nkernel: " + str(kernel)
        d.add_history(history)
    return new_data


def get_kernel(kernel, default=moments):
    """
     Return a valid kernel instance from either an instance or a string, default is moments if no kernel is specified

    :param str or cis.collocation.col_framework.Kernel kernel:
    :param default:
    :return cis.collocation.col_framework.Kernel:
    """
    from cis.collocation.col_framework import get_kernel, Kernel
    if not kernel:
        kernel = default()
    elif isinstance(kernel, six.string_types):
        kernel = get_kernel(kernel)()
    elif not isinstance(kernel, Kernel):
        raise ValueError("Invalid kernel argument, it must be either a string or a Kernel instance")
    return kernel



"""
Module to do integration tests of plots to files. Does not check the graph created is correct, only that a graph is
created without errors.
"""
from cis.exceptions import UserPrintableException
from cis.cis_main import plot_cmd, subset_cmd, aggregate_cmd
from cis.parse import parse_args
from cis.test.integration_test_data import *
from cis.test.integration.base_integration_test import BaseIntegrationTest
from nose.tools import raises


class TestPlotIntegration(BaseIntegrationTest):

    def test_should_do_scatter_plot_of_file_valid_aerosol_cci_file(self):
        # Actual file name: 20080612093821-ESACCI-L2P_AEROSOL-ALL-AATSR_ENVISAT-ORAC_32855-fv02.02.nc
        arguments = ['plot', valid_aerosol_cci_variable+':'+escape_colons(valid_aerosol_cci_filename), '--output',
                     valid_aerosol_cci_filename+'.png']
        main_arguments = parse_args(arguments)
        plot_cmd(main_arguments)

        # Remove plotted file, will throw an OSError if file was not created
        os.remove(valid_aerosol_cci_filename+'.png')

    def test_should_do_scatter_plot_of_valid_2d_file(self):
        # Actual file name: xglnwa.pm.k8dec-k9nov.col.tm.nc
        arguments = ['plot', valid_1d_variable+':'+escape_colons(valid_2d_filename), '--output', valid_2d_filename+'.png']
        main_arguments = parse_args(arguments)
        plot_cmd(main_arguments)

        # Remove plotted file, will throw an OSError if file was not created
        os.remove(valid_2d_filename+'.png')

    def test_should_do_scatter_plot_of_valid_2d_file_with_x_as_lat_y_as_long(self):
        # Actual file name: xglnwa.pm.k8dec-k9nov.col.tm.nc
        # Trac issue #153
        arguments = ['plot', valid_1d_variable+':'+escape_colons(valid_2d_filename), '--xaxis', 'latitude', '--yaxis', 'longitude',
                     '--output', valid_2d_filename+'.png']
        main_arguments = parse_args(arguments)
        plot_cmd(main_arguments)

        # Remove plotted file, will throw an OSError if file was not created
        os.remove(valid_2d_filename+'.png')

    def test_should_do_heatmap_plot_of_valid_1d_file(self):
        # Actual file name: xglnwa.pm.k8dec-k9nov.vprof.tm.nc
        # Trac issue #145
        arguments = ['plot', valid_1d_variable+':'+escape_colons(valid_1d_filename), '--output',
                     valid_1d_filename+'.png']
        main_arguments = parse_args(arguments)
        plot_cmd(main_arguments)

        # Remove plotted file, will throw an OSError if file was not created
        os.remove(valid_1d_filename+'.png')

    def test_should_do_line_plot_of_valid_zonal_time_mean_cmip5_file(self):
        # Actual file name: rsutcs_Amon_HadGEM2-A_sstClim_r1i1p1_185912-188911.CMIP5.tm.zm.nc
        # Trac issue #130
        arguments = ['plot',
                     valid_zonal_time_mean_CMIP5_variable+':'+escape_colons(valid_zonal_time_mean_CMIP5_filename),
                     '--output', valid_zonal_time_mean_CMIP5_filename+'.png']
        main_arguments = parse_args(arguments)
        plot_cmd(main_arguments)

        # Remove plotted file, will throw an OSError if file was not created
        os.remove(valid_zonal_time_mean_CMIP5_filename+'.png')

    def test_should_do_plot_of_hybrid_height_when_formula_terms_not_marked_as_coordinates(self):
        # Actual file name: hybrid-height.nc
        # Trac issue #417 (fixed in IRIS v.1.7.1)
        arguments = ['plot', valid_hybrid_height_flat_variable+':'+escape_colons(valid_hybrid_height_flat_filename),
                     '--output', valid_hybrid_height_filename+'.png']
        main_arguments = parse_args(arguments)
        plot_cmd(main_arguments)

        # Remove plotted file, will throw an OSError if file was not created
        os.remove(valid_hybrid_height_filename+'.png')

    def test_should_do_plot_of_hybrid_pressure_using_calculated_air_pressure_variable(self):
        # Actual file name: hybrid-pressure.nc
        # Trac issue #416 (fixed in IRIS v.1.7.1)
        arguments = ['plot', valid_hybrid_pressure_variable+':'+escape_colons(valid_hybrid_pressure_filename),
                     '--xaxis', 'air_pressure', '--output', valid_hybrid_pressure_filename+'.png']
        main_arguments = parse_args(arguments)
        plot_cmd(main_arguments)

        # Remove plotted file, will throw an OSError if file was not created
        os.remove(valid_hybrid_pressure_filename+'.png')

    def test_subset_ECHAM_over_0_360_boundary_plots_OK(self):
        var = valid_echamham_variable_1
        filename = valid_echamham_filename
        args = ['subset', var + ':' + escape_colons(filename) + ":product=NetCDF_Gridded", 'x=[-10,10]', '-o', self.OUTPUT_FILENAME]
        args = parse_args(args)
        subset_cmd(args)
        out_name = 'subset_echam_boundary.png'
        args = ['plot', var + ":" + self.OUTPUT_FILENAME + ":product=NetCDF_Gridded", '--type', 'contourf', '-o', out_name]
        args = parse_args(args)
        plot_cmd(args)

        os.remove(out_name)

    @raises(ValueError)
    def test_plot_gridded_3d_exits_with_ValueError(self):
        variable = valid_cis_gridded_output_variable
        filename = valid_cis_gridded_output_filename
        out_name = '3d_out.png'
        args = ['plot', variable + ':' + escape_colons(filename) + ":product=NetCDF_Gridded", '-o', out_name]
        main_arguments = parse_args(args)
        plot_cmd(main_arguments)

    def test_plot_gridded_2d_with_flattened_time(self):
        variable = valid_cis_gridded_output_variable
        filename = valid_cis_gridded_output_filename
        args = ['subset', variable + ':' + escape_colons(filename) + ":product=NetCDF_Gridded",
                't=[2007-06-07T15]', '-o', self.OUTPUT_FILENAME]
        args = parse_args(args)
        subset_cmd(args)
        out_name = '3d_out.png'
        args = ['plot', variable + ':' + self.OUTPUT_FILENAME + ":product=NetCDF_Gridded", '-o', out_name]
        main_arguments = parse_args(args)
        plot_cmd(main_arguments)

        os.remove(out_name)

    def test_plot_aggregated_aeronet(self):
        # Aggregated aeronet has multiple length 1 dimensions, so we want to make sure we can plot it OK.
        # JASCIS-183
        variable = 'Solar_Zenith_Angle'
        filename = valid_aeronet_filename
        agg_args = ['aggregate', variable + ':' + escape_colons(filename),
                    't=[2003-09-24T07:00:00,2003-11-04T07:00:00,P1D]', '-o', self.OUTPUT_FILENAME]
        args = parse_args(agg_args)
        aggregate_cmd(args)
        out_name = 'aeronet_out.png'
        args = ['plot', variable + ':' + self.OUTPUT_FILENAME,
                '--xaxis', 'time', '--yaxis', variable, '-o', out_name]
        main_arguments = parse_args(args)
        plot_cmd(main_arguments)

        os.remove(out_name)

    def test_plot_ungridded_histogram(self):
        filename = valid_GASSP_station_filename
        variable = valid_GASSP_station_vars[0]
        out_name = 'histogram.png'
        args = ['plot', variable + ':' + escape_colons(filename), '--type', 'histogram', '-o', out_name]
        args = parse_args(args)
        plot_cmd(args)

        os.remove(out_name)

    @raises(UserPrintableException)
    def test_plot_ungridded_heatmap(self):
        filename = valid_GASSP_station_filename
        variable = valid_GASSP_station_vars[0]
        args = ['plot', variable + ':' + escape_colons(filename), '--type', 'heatmap']
        args = parse_args(args)
        plot_cmd(args)

    def test_plot_heatmap_horizontal_cbar(self):
        var = valid_echamham_variable_1
        filename = valid_echamham_filename
        out_name = 'cbarh.png'
        args = ['plot', var + ':' + escape_colons(filename), '--cbarorient', 'horizontal', '-o', out_name]
        args = parse_args(args)
        plot_cmd(args)

        os.remove(out_name)

    def test_plot_heatmap_vertical_cbar(self):
        var = valid_echamham_variable_1
        filename = valid_echamham_filename
        out_name = 'cbarv.png'
        args = ['plot', var + ':' + escape_colons(filename), '--cbarorient', 'vertical', '-o', out_name]
        args = parse_args(args)
        plot_cmd(args)

        os.remove(out_name)

    def test_plot_heatmap_cbar_scale(self):
        var = valid_echamham_variable_1
        filename = valid_echamham_filename
        out_name = 'cbarscale.png'
        args = ['plot', var + ':' + escape_colons(filename), '--cbarscale', '0.75', '-o', out_name]
        args = parse_args(args)
        plot_cmd(args)

        os.remove(out_name)



import unittest

import numpy as np
from hamcrest import assert_that, is_, close_to

from cis.stats import StatsAnalyzer
from cis.data_io.gridded_data import GriddedData
from cis.test.util import mock


class TestStatsAnalyser(unittest.TestCase):

    def setUp(self):
        self.data1 = GriddedData(np.array([[1.0, -2.0, 3.0, 4.0, 5.0],
                                           [6.0,  7.0, 8.0, 9.0, 10.0]]))
        self.data2 = GriddedData(np.array([[0.8, -1.7, 3.6, 3.7, 5.1],
                                           [6.0,  7.3, 7.9, 9.1, 10.1]]))
        self.missing1 = GriddedData(np.ma.masked_equal([[1.0, -2.0, 3.0, 4.0, -999],
                                                        [6.0, -999, 8.0, -100, -999]], -999))
        self.missing2 = GriddedData(np.ma.masked_equal([[-999, -2.1, 4.1, 3.2, -999],
                                                        [6.0, 7.3,  8.1, -120, -999]], -999))

    def test_GIVEN_datasets_WHEN_analyze_THEN_StatisticsResults_returned(self):
        stats = StatsAnalyzer(self.data1, self.data2)
        results = stats.analyze()
        assert_that(len(results), is_(14))

    def test_GIVEN_flattened_and_unflattened_datasets_WHEN_analyze_THEN_StatisticsResults_returned(self):
        data1 = mock.make_regular_2d_ungridded_data()
        data2 = mock.make_regular_2d_ungridded_data()
        data2._data = data2.data_flattened
        for coord in data2.coords():
            coord._data = coord.data_flattened
        stats = StatsAnalyzer(data1, data2)
        results = stats.analyze()
        assert_that(len(results), is_(14))

    def test_GIVEN_missing_values_WHEN_analyze_THEN_original_data_unchanged(self):
        # We perform some manipulation on the data masks, but we don't want the
        # original data to be changed.
        stats = StatsAnalyzer(self.missing1, self.missing2)
        results = stats.analyze()
        assert_that(len(self.missing1.data.compressed()), is_(7))
        assert_that(len(self.missing2.data.compressed()), is_(7))

    # ======================== POINTS COUNT

    def test_GIVEN_no_missing_vals_WHEN_count_THEN_points_count_correct(self):
        stats = StatsAnalyzer(self.data1, self.data2)
        res = stats.points_count()
        assert_that(res[0].num_points, is_(10))

    def test_GIVEN_missing_vals_WHEN_count_THEN_points_count_correct(self):
        stats = StatsAnalyzer(self.missing1, self.missing2)
        res = stats.points_count()
        assert_that(res[0].num_points, is_(6))

    def test_GIVEN_one_masked_one_nparray_WHEN_points_count_THEN_count_correct(self):
        stats = StatsAnalyzer(self.data1, self.missing2)
        res = stats.points_count()
        assert_that(res[0].num_points, is_(7))

    # ======================== MEAN

    def test_GIVEN_no_missing_vals_WHEN_mean_THEN_mean_correct(self):
        stats = StatsAnalyzer(self.data1, self.data2)
        res = stats.means()
        assert_that(res[0].mean, close_to(5.1, 1e-5))
        assert_that(res[1].mean, close_to(5.19, 1e-5))

    def test_GIVEN_missing_vals_WHEN_count_THEN_mean_correct(self):
        stats = StatsAnalyzer(self.missing1, self.missing2)
        res = stats.means()
        assert_that(res[0].mean, close_to(-13.5, 1e-5))
        assert_that(res[1].mean, close_to(-16.783333333, 1e-5))

    # ======================= STDDEV

    def test_GIVEN_no_missing_vals_WHEN_stddev_THEN_stddev_correct(self):
        stats = StatsAnalyzer(self.data1, self.data2)
        res = stats.stddevs()
        assert_that(res[0].stddev, close_to(3.7252889523, 1e-5))
        assert_that(res[1].stddev, close_to(3.7020864988, 1e-5))

    def test_GIVEN_missing_vals_WHEN_stddev_THEN_stddev_correct(self):
        stats = StatsAnalyzer(self.missing1, self.missing2)
        res = stats.stddevs()
        assert_that(res[0].stddev, close_to(42.5099988238, 1e-5))
        assert_that(res[1].stddev, close_to(50.6813344997, 1e-5))

    # ================= ABS MEAN

    def test_GIVEN_no_missing_vals_WHEN_abs_mean_THEN_abs_mean_correct(self):
        stats = StatsAnalyzer(self.data1, self.data2)
        res = stats.abs_mean()
        assert_that(res[0].abs_mean, close_to(0.09, 1e-5))

    def test_GIVEN_missing_vals_WHEN_abs_mean_THEN_abs_mean_correct(self):
        stats = StatsAnalyzer(self.missing1, self.missing2)
        res = stats.abs_mean()
        assert_that(res[0].abs_mean, close_to(-3.2833333333, 1e-5))

    def test_GIVEN_one_masked_one_nparray_WHEN_abs_mean_THEN_abs_mean_correct(self):
        stats = StatsAnalyzer(self.data1, self.missing2)
        res = stats.abs_mean()
        assert_that(res[0].abs_mean, close_to(-18.3428571429, 1e-5))

    # ================= ABS STDDEV

    def test_GIVEN_no_missing_vals_WHEN_abs_stddev_THEN_abs_stddev_correct(self):
        stats = StatsAnalyzer(self.data1, self.data2)
        res = stats.abs_stddev()
        assert_that(res[0].abs_stddev, close_to(0.2643650675, 1e-5))

    def test_GIVEN_missing_vals_WHEN_abs_stddev_THEN_abs_stddev_correct(self):
        stats = StatsAnalyzer(self.missing1, self.missing2)
        res = stats.abs_stddev()
        assert_that(res[0].abs_stddev, close_to(8.2120440005, 1e-5))

    def test_GIVEN_one_masked_one_nparray_WHEN_abs_stddev_THEN_abs_stddev_correct(self):
        stats = StatsAnalyzer(self.data1, self.missing2)
        res = stats.abs_stddev()
        assert_that(res[0].abs_stddev, close_to(48.7984582114, 1e-5))

    # ================= REL MEAN
    
    def test_GIVEN_no_missing_vals_WHEN_rel_mean_THEN_rel_mean_correct(self):
        stats = StatsAnalyzer(self.data1, self.data2)
        res = stats.rel_mean()
        assert_that(res[0].rel_mean, close_to(-0.0153531746, 1e-5))

    def test_GIVEN_missing_vals_WHEN_rel_mean_THEN_rel_mean_correct(self):
        stats = StatsAnalyzer(self.missing1, self.missing2)
        res = stats.rel_mean()
        assert_that(res[0].rel_mean, close_to(0.0715277778, 1e-5))

    def test_GIVEN_one_masked_one_nparray_WHEN_rel_mean_THEN_rel_mean_correct(self):
        stats = StatsAnalyzer(self.data1, self.missing2)
        res = stats.rel_mean()
        assert_that(res[0].rel_mean, close_to(-2.0087585034, 1e-5))
    
    def test_GIVEN_divide_by_zeros_WHEN_rel_mean_THEN_rel_mean_correct(self):
        # Zeros in the first array cause undefined values in the relative calculations.
        data_with_zeros = self.data1
        data_with_zeros.data[0][1] = 0
        stats = StatsAnalyzer(data_with_zeros, self.missing2)
        res = stats.rel_mean()
        assert_that(res[0].rel_mean, close_to(-2.3518849206, 1e-5))

    # ================= REL STDDEV
    
    def test_GIVEN_no_missing_vals_WHEN_rel_stddev_THEN_rel_stddev_correct(self):
        stats = StatsAnalyzer(self.data1, self.data2)
        res = stats.rel_stddev()
        assert_that(res[0].rel_stddev, close_to(0.1097392069, 1e-5))

    def test_GIVEN_missing_vals_WHEN_rel_stddev_THEN_rel_stddev_correct(self):
        stats = StatsAnalyzer(self.missing1, self.missing2)
        res = stats.rel_stddev()
        assert_that(res[0].rel_stddev, close_to(0.1930820326, 1e-5))

    def test_GIVEN_one_masked_one_nparray_WHEN_rel_stddev_THEN_rel_stddev_correct(self):
        stats = StatsAnalyzer(self.data1, self.missing2)
        res = stats.rel_stddev()
        assert_that(res[0].rel_stddev, close_to(5.4371807462, 1e-5))

    def test_GIVEN_divide_by_zeros_WHEN_rel_stddev_THEN_rel_stddev_correct(self):
        # Zeros in the first array cause undefined values in the relative calculations.
        data_with_zeros = self.data1
        data_with_zeros.data[0][1] = 0
        stats = StatsAnalyzer(data_with_zeros, self.missing2)
        res = stats.rel_stddev()
        assert_that(res[0].rel_stddev, close_to(5.8725242578, 1e-5))

    # ================= SPEARMAN

    def test_GIVEN_no_missing_vals_WHEN_spearman_THEN_spearman_correct(self):
        stats = StatsAnalyzer(self.data1, self.data2)
        res = stats.spearmans_rank()
        assert_that(res[0].spearman, close_to(1.0, 1e-5))

    def test_GIVEN_missing_vals_WHEN_spearman_THEN_spearman_correct(self):
        stats = StatsAnalyzer(self.missing1, self.missing2)
        res = stats.spearmans_rank()
        assert_that(res[0].spearman, close_to(0.9428571429, 1e-5))

    def test_GIVEN_one_masked_one_nparray_WHEN_spearman_THEN_spearman_correct(self):
        stats = StatsAnalyzer(self.data1, self.missing2)
        res = stats.spearmans_rank()
        assert_that(res[0].spearman, close_to(0.2142857143, 1e-5))

    # ==================  LINREG

    def test_GIVEN_no_missing_vals_WHEN_lin_regression_THEN_regression_correct(self):
        stats = StatsAnalyzer(self.data1, self.data2)
        res = stats.linear_regression()
        expected_res = [0.9912730184, 0.1345076061, 0.997485722,  0.0248994694107]
        actual_res = res[0].grad, res[1].intercept, res[2].r, res[3].stderr
        assert_that(np.allclose(actual_res, expected_res))

    def test_GIVEN_missing_vals_WHEN_lin_regression_THEN_regression_correct(self):
        stats = StatsAnalyzer(self.missing1, self.missing2)
        res = stats.linear_regression()
        expected_res = [1.1920369653, -0.6908343017, 0.999845219, 0.0104877890357]
        actual_res = res[0].grad, res[1].intercept, res[2].r, res[3].stderr
        assert_that(np.allclose(actual_res, expected_res))

    def test_GIVEN_one_masked_one_nparray_WHEN_lin_regression_THEN_regression_correct(self):
        stats = StatsAnalyzer(self.data1, self.missing2)
        res = stats.linear_regression()
        expected_res = [-5.1404761905, 12.3595238095, -0.4079085869, 5.14561290806]
        actual_res = res[0].grad, res[1].intercept, res[2].r, res[3].stderr
        assert_that(np.allclose(actual_res, expected_res))

if __name__ == '__main__':
    unittest.main()



"""
Indexes over data used for fast lookup when collocating.
"""
import logging
import math
import datetime

import numpy as np
import numpy.ma as ma

from cis.collocation.haversinedistancekdtreeindex import HaversineDistanceKDTreeIndex
from cis.time_util import convert_datetime_to_std_time


class GridCellBinIndexInSlices(object):
    def __init__(self):
        # cells numbers for each hyperpoint
        self.cell_numbers = None

        # sort order for each hyperpoint, i.e. where they are
        self.sort_order = None

        # position of each hyperpoint in the grid (coordinate x point)
        self._indices = None

        # indexes for the slices through the sorted data.
        # e.g. cell_slices_index[0] = [0, 2] would be points 0, 1 and 2 are in the first cell
        self.cell_slices_indices = None

        # hyper point cordinates list
        self.hp_coords = None

    def index_data(self, coords, hyper_points, coord_map):
        """
        Index the data that falls inside the grid cells

        :param coords: coordinates of grid
        :param hyper_points: list of HyperPoints to index
        :param coord_map: list of tuples relating index in HyperPoint to index in coords and in
                          coords to be iterated over
        """

        # create bounds in correct order
        hp_coords = []
        coord_descreasing = [False] * len(coords)
        coord_lengths = [0] * len(coords)
        lower_bounds = [None] * len(coords)
        max_bounds = [None] * len(coords)
        for (hpi, ci, shi) in coord_map:
            coord = coords[ci]
            # Coordinates must be monotonic; determine whether increasing or decreasing.
            if len(coord.points) > 1:
                if coord.points[1] < coord.points[0]:
                    coord_descreasing[shi] = True
            coord_lengths[shi] = len(coord.points)
            if coord_descreasing[shi]:
                lower_bounds[shi] = coord.bounds[::-1, 1]
                max_bounds[shi] = coord.bounds[0, 1]
            else:
                lower_bounds[shi] = coord.bounds[::, 0]
                max_bounds[shi] = coord.bounds[-1, 1]

            hp_coord = hyper_points.coords[hpi]
            if isinstance(hp_coord[0], datetime.datetime):
                hp_coord = convert_datetime_to_std_time(hp_coord)

            hp_coords.append(hp_coord)

        bounds_coords_max = list(zip(lower_bounds, hp_coords, max_bounds))

        # stack for each coordinate
        #    where the coordinate is larger than the maximum set to -1
        #    otherwise search in the sorted coordinate to find all the index of the hyperpoints
        # The choice of 'left' or 'right' and '<' and '<=' determines which
        #  cell is chosen when the coordinate is equal to the boundary.
        # -1 or M_i indicates the point is outside the grid.
        # Output is a list of coordinates which lists the indexes where the hyper points
        #    should be located in the grid
        indices = np.vstack(
            np.where(
                ci < max_coordinate_value,
                np.searchsorted(bi, ci, side='right') - 1,
                -1)
            for bi, ci, max_coordinate_value in bounds_coords_max)

        # D-tuple giving the shape of the output grid
        grid_shape = tuple(len(bi_ci[0]) for bi_ci in bounds_coords_max)

        # shape (N,) telling which points actually fall within the grid,
        # i.e. have indexes that are not -1 and are not masked data points
        grid_mask = np.all(
            (indices >= 0) &
            (ma.getmaskarray(hyper_points.data) == False),
            axis=0)

        # if the coordinate was decreasing then correct the indices for this cell
        for indices_slice, decreasing, coord_length in zip(range(indices.shape[0]), coord_descreasing, coord_lengths):
            if decreasing:
                # indices[indices_slice] += (coord_length - 1) - indices[indices_slice]
                indices[indices_slice] *= -1
                indices[indices_slice] += (coord_length - 1)

        # shape (N,) containing negative scalar cell numbers for each
        # input point (sequence doesn't matter so long as they are unique), or
        # -1 for points outside the grid.
        #
        # Possibly numpy.lexsort could be used to avoid the need for this,
        # although we'd have to be careful about points outside the grid.
        self.cell_numbers = np.where(
            grid_mask,
            np.tensordot(
                np.cumproduct((1,) + grid_shape[:-1]),
                indices,
                axes=1
            ),
            -1)

        # Sort everything by cell number
        self.sort_order = np.argsort(self.cell_numbers)
        self.cell_numbers = self.cell_numbers[self.sort_order]
        self._indices = indices[:, self.sort_order]
        self.hp_coords = [hp_coord[self.sort_order] for hp_coord in hp_coords]

    def get_iterator(self):
        """
        Get an iterator through all the points which will contribute to a cell.
        Iteration is through out indices (where the data point is in the grid) and the
        (start, stop) indexes in a sorted list of the points in that cell.
        self.sort_order can be used to order the list

        :return: an iterator out_indices, cell_slice_indices
        """

        # find the index at which the cell number changes, +1 to make this the first point with the new
        # cell number
        indexes_of_first_element_in_slice = np.flatnonzero(np.diff(self.cell_numbers)) + 1

        # shape (L,2) giving the pairs (first, last+1) of indices such that
        # cell_numbers[first:last+1] is the slice of all the elements of each
        # of the L unique values.

        # if some of the points are not in the grid skip these
        if self.cell_numbers[0] == -1:
            # if there are no points in the grid (all -1s)
            if indexes_of_first_element_in_slice.size == 0:
                self.cell_slices_indices = []
            else:
                # for list of start stop indexes starting at first point in the grid
                self.cell_slices_indices = np.concatenate((
                    indexes_of_first_element_in_slice[:1],
                    np.tile(indexes_of_first_element_in_slice[1:], 2),
                    [len(self.cell_numbers)]
                )).reshape(2, -1).T
        else:
            # create list of start stop indexes starting at 0
            self.cell_slices_indices = np.concatenate((
                [0],  # start at first element
                np.tile(indexes_of_first_element_in_slice, 2),
                [len(self.cell_numbers)]  # last at first element
            )).reshape(2, -1).T  # reshape so that it is list of start-end indices

        # iterate around slices
        for cell_slice_indices in self.cell_slices_indices:
            out_indices = tuple(self._indices[:, cell_slice_indices[0]])
            yield out_indices, cell_slice_indices


class GridCellBinIndex(object):
    def __init__(self):
        self.index = None

    def index_data(self, coords, data, coord_map):
        """
        Creates an index of points that fall within grid cells.

        :param coords: coordinates of grid
        :param data: list of HyperPoints to index
        :param coord_map: list of tuples relating index in HyperPoint to index in coords and in
                          coords to be iterated over
        """
        # Create an index array matching the shape of the coordinates to be iterated over.
        shape = []
        for (hpi, ci, shi) in coord_map:
            shape.append(len(coords[ci].points))
        num_cell_indices = len(shape)

        # Set a logging interval.
        num_bin_checks = sum([math.log(x) for x in shape])
        log_every_points = 2000000 / num_bin_checks
        log_every_points -= log_every_points % 100
        log_every_points = max(log_every_points, 100)

        # Create the index, which will be an array containing of a lists of data points that are
        # within each cell.
        self.index = np.empty(tuple(shape), dtype=np.dtype(object))
        self.index.fill(None)

        coord_decreasing = [False] * len(coords)
        coord_lengths = [0] * len(coords)
        lower_bounds = [None] * len(coords)
        upper_bounds = [None] * len(coords)
        for (hpi, ci, shi) in coord_map:
            coord = coords[ci]
            # Coordinates must be monotonic; determine whether increasing or decreasing.
            if len(coord.points) > 1:
                if coord.points[1] < coord.points[0]:
                    coord_decreasing[ci] = True
            coord_lengths[ci] = len(coord.points)
            if coord_decreasing[ci]:
                lower_bounds[ci] = coord.bounds[::-1, 1]
                upper_bounds[ci] = coord.bounds[::-1, 0]
            else:
                lower_bounds[ci] = coord.bounds[::, 0]
                upper_bounds[ci] = coord.bounds[::, 1]

        # Populate the index by finding the cell containing each data point.
        num_points = len(data)
        pt_count = 0
        pt_total = 0
        for pt_idx, point in data.enumerate_non_masked_points():
            point_cell_indices = [None] * num_cell_indices

            # Find the interval that the point resides in for each relevant coordinate.
            for (hpi, ci, shi) in coord_map:
                search_index = np.searchsorted(lower_bounds[ci], point[hpi], side='right') - 1
                if (search_index >= 0) and (point[hpi] < upper_bounds[ci][search_index]):
                    if coord_decreasing[ci]:
                        point_cell_indices[shi] = coord_lengths[ci] - search_index - 1
                    else:
                        point_cell_indices[shi] = search_index

            # Add point to index if a containing interval was found for each coordinate.
            if point_cell_indices.count(None) == 0:
                index_entry = self.index[tuple(point_cell_indices)]
                if index_entry is None:
                    index_entry = []
                index_entry.append(pt_idx)
                self.index[tuple(point_cell_indices)] = index_entry

            # Periodically log progress.
            pt_count += 1
            pt_total += 1
            if pt_count == log_every_points:
                logging.info("    Indexed %d points of %d (%d%%)",
                             pt_total, num_points, int(pt_total * 100 / num_points))
                pt_count = 0

    def get_points_by_indices(self, indices):
        return self.index[tuple(indices)]


# Map of names of attributes of a constraint or kernel to the class used to
# create an index to which the attribute should be set
_index_attributes = {'grid_cell_bin_index': GridCellBinIndex,
                     'grid_cell_bin_index_slices': GridCellBinIndexInSlices,
                     'haversine_distance_kd_tree_index': HaversineDistanceKDTreeIndex}


def create_indexes(operator, coords, data, coord_map):
    """
    :param operator: constraint or kernel instance
    :param coords: coordinates of grid
    :param data: list of HyperPoints to index
    :param coord_map: list of tuples relating index in HyperPoint to index in coords and in
                      coords to be iterated over
    """
    for attr, cls in _index_attributes.items():
        if hasattr(operator, attr):
            index = cls()
            logging.info("--> Creating index for %s", operator.__class__.__name__)
            index.index_data(coords, data, coord_map)
            setattr(operator, attr, index)



"""
module to test the various subclasses of the abstract AProduct class
"""

import unittest
import logging
import os
from tempfile import mkdtemp

from hamcrest import *
from nose.tools import istest, eq_, raises, nottest, with_setup, assert_is_instance

from cis.data_io.products import *
from cis.exceptions import InvalidVariableError, FileFormatError
from cis.test.integration_test_data import non_netcdf_file, cis_test_files, skip_pyhdf
from cis.data_io.products import CloudSat

try:
    import pandas
except ImportError:
    # Disable all these tests if pandas is not installed.
    pandas = None

skip_pandas = unittest.skipIf(pandas is None, 'Test(s) require "pandas", which is not available.')


def check_regex_matching(cls_name, filename):
    from cis.data_io.products.AProduct import __get_class
    cls = __get_class(filename)
    eq_(cls.__name__, cls_name)

invalid_variable = "im_an_invalid_variable"
invalid_filename = "im_an_invalid_file"
invalid_format = non_netcdf_file

test_file = os.path.join(mkdtemp('cis_test_dir'), 'test_out.nc')


def remove_test_file():
    """
        Try and remove the test file if it exists
    :return:
    """
    from os import remove
    from os.path import isfile
    if isfile(test_file):
        remove(test_file)


class ProductTests(object):
    product = None
    filename = None
    vars = None
    valid_variable = None
    file_format = None
    test_file_metadata = None
    valid_vars_count = None

    @nottest
    def setup(self, test_file_metadata, product):
        from cis.test.integration_test_data import TestFileTestData
        assert isinstance(test_file_metadata, TestFileTestData)
        self.filename = test_file_metadata.master_filename
        self.valid_variable = test_file_metadata.data_variable_name
        self.vars = test_file_metadata.all_variable_names
        self.valid_vars_count = test_file_metadata.valid_vars_count
        self.file_format = test_file_metadata.file_format
        self.test_file_metadata = test_file_metadata
        self.product = product

    @istest
    def test_file_regex_matching(self):
        check_regex_matching(self.product.__name__, self.filename)

    @istest
    def test_file_regex_matching_for_full_path(self):
        import os
        check_regex_matching(self.product.__name__,  os.path.join(os.getcwd(), self.filename))

    @istest
    def test_variable_wildcarding(self):
        vars = self.product().get_variable_names([self.filename])
        self.check_valid_vars(vars)

    @istest
    def test_variable_printing(self):
        import six
        data_obj = self.product().create_data_object([self.filename], self.valid_variable)
        summary = data_obj.summary()
        assert_is_instance(summary, six.string_types)

    def check_valid_vars(self, vars):
        if self.vars is not None:
            assert_that(set(vars), is_(set(self.vars)), "Variables")
        else:
            assert_that(len(vars), is_(self.valid_vars_count), "Number of valid variables in the file")

    def test_create_data_object(self):
        from cis.data_io.common_data import CommonData
        data = self.product().create_data_object([self.filename], self.valid_variable)
        assert_is_instance(data, CommonData)

    @skip_pandas
    def test_create_pandas_object(self):
        from pandas import DataFrame
        data = self.product().create_data_object([self.filename], self.valid_variable)
        df = data.as_data_frame()
        assert_is_instance(df, DataFrame)

    def test_create_coords(self):
        valid_standard_names = ['latitude', 'longitude', 'altitude', 'time', 'air_pressure']
        coords = self.product().create_coords([self.filename])
        coord_list = coords.coords()

        for coord in coord_list:
            logging.debug(coord.standard_name)

        assert (all([coord.standard_name in valid_standard_names for coord in coord_list]))

    @with_setup(remove_test_file, remove_test_file)
    def test_write_coords(self):
        from cis.data_io.write_netcdf import write_coordinates
        coords = self.product().create_coords([self.filename], self.valid_variable)
        write_coordinates(coords, test_file)

    @raises(IOError)
    def test_should_raise_ioerror_with_invalid_filename(self):
        self.product().create_data_object([invalid_filename], self.valid_variable)

    @raises(IOError, FileFormatError)
    def test_should_raise_ioerror_or_translationerror_with_file_that_is_not_a_recognised_format(self):
        self.product().create_data_object([invalid_format], self.valid_variable)

    @raises(InvalidVariableError)
    def test_should_raise_error_when_variable_does_not_exist_in_file(self):
        self.product().create_data_object([self.filename], invalid_variable)

    def test_file_format(self):
        expected_file_format = self.file_format
        if expected_file_format is None:
            expected_file_format = self.product.__name__
        file_format = self.product().get_file_format(self.filename)
        assert_that(file_format, is_(expected_file_format), "File format")


@skip_pyhdf
class TestCloudsatRVODsdata(ProductTests, unittest.TestCase):
    def setUp(self):
        from cis.test.integration_test_data import valid_cloudsat_RVOD_file, valid_cloudsat_RVOD_sdata_variable, \
            valid_cloudsat_RVOD_file_format

        self.filename = valid_cloudsat_RVOD_file
        self.valid_variable = valid_cloudsat_RVOD_sdata_variable
        self.product = CloudSat
        self.file_format = valid_cloudsat_RVOD_file_format

    def check_valid_vars(self, vars):
        assert len(vars) == 47 + 22  # Expect 47 valid SD vars and 22 valid VD vars


@skip_pyhdf
class TestCloudsatRVODvdata(ProductTests, unittest.TestCase):
    def setUp(self):
        from cis.test.integration_test_data import valid_cloudsat_RVOD_file, valid_cloudsat_RVOD_vdata_variable, \
            valid_cloudsat_RVOD_file_format

        self.filename = valid_cloudsat_RVOD_file
        self.valid_variable = valid_cloudsat_RVOD_vdata_variable
        self.product = CloudSat
        self.file_format = valid_cloudsat_RVOD_file_format

    def check_valid_vars(self, vars):
        assert len(vars) == 47 + 22  # Expect 47 valid SD vars and 22 valid VD vars


@skip_pyhdf
class TestCloudsatPRECIP(ProductTests, unittest.TestCase):
    def setUp(self):
        from cis.test.integration_test_data import valid_cloudsat_PRECIP_file, valid_cloudsat_PRECIP_variable, \
            valid_cloudsat_PRECIP_file_format

        self.filename = valid_cloudsat_PRECIP_file
        self.valid_variable = valid_cloudsat_PRECIP_variable
        self.product = CloudSat
        self.file_format = valid_cloudsat_PRECIP_file_format
        self.vars = ['Profile_time',
                     'Latitude',
                     'Longitude',
                     'DEM_elevation',
                     'Data_quality',
                     'Data_status',
                     'Data_targetID',
                     'Precip_rate',
                     'Precip_rate_min',
                     'Precip_rate_max',
                     'Precip_rate_no_ms',
                     'PIA_hydrometeor',
                     'PIA_near_surface',
                     'Sigma_zero',
                     'Near_surface_reflectivity',
                     'Cloud_top_lowest_layer',
                     'Cloud_top_highest_layer',
                     'Frozen_precip_height',
                     'Melted_fraction',
                     'Status_flag',
                     'Precip_flag',
                     'Retrieval_info_flag',
                     'Phase_flag',
                     'Cloud_flag',
                     'SRT_flag',
                     'Surface_type',
                     'Freezing_level',
                     'SST',
                     'Surface_wind',
                     'Aux_CWV_AMSR',
                     'Aux_LWP_AMSR',
                     'Aux_SST_AMSR',
                     'Aux_precip_AMSR',
                     'Aux_dist_AMSR']



@skip_pyhdf
class TestCaliop_L2(ProductTests, unittest.TestCase):
    def setUp(self):
        self.setup(cis_test_files["caliop_L2"], Caliop_L2)


@skip_pyhdf
class TestCaliop_L1(ProductTests, unittest.TestCase):
    def setUp(self):
        self.setup(cis_test_files["caliop_L1"], Caliop_L1)


@skip_pyhdf
class TestMODIS_L3(ProductTests, unittest.TestCase):
    def setUp(self):
        self.setup(cis_test_files["modis_L3"], MODIS_L3)

    def check_valid_vars(self, vars):
        assert len(vars) == 700

    @nottest
    def test_write_coords(self):
        # Model data is gridded so IRIS takes care of this
        pass

    @nottest
    def test_create_pandas_object(self):
        # Cannot create pandas objects for data with > 2 dimensions
        pass


@skip_pyhdf
class TestMODIS_L2(ProductTests, unittest.TestCase):
    def setUp(self):

        self.setup(cis_test_files["modis_L2"], MODIS_L2)


class TestCloud_CCI(ProductTests, unittest.TestCase):
    def setUp(self):
        self.setup(cis_test_files["Cloud_CCI"], Cloud_CCI_L2)


class TestAerosol_CCI(ProductTests, unittest.TestCase):
    def setUp(self):
        self.setup(cis_test_files["Aerosol_CCI"], Aerosol_CCI_L2)

    def check_valid_vars(self, vars):
        exclude_vars = ["sun_zenith", "satellite_zenith", "relative_azimuth", "instrument_view"]
        assert len(vars) == 32
        for key in exclude_vars:
            assert key not in vars


class TestCis(ProductTests, unittest.TestCase):
    def setUp(self):
        self.setup(cis_test_files["CIS_Ungridded"], cis)


class TestAeronet(ProductTests, unittest.TestCase):
    def setUp(self):
        from cis.test.integration_test_data import valid_aeronet_filename, another_valid_aeronet_filename
        self.setup(cis_test_files["aeronet"], Aeronet)
        self.filenames = [valid_aeronet_filename, another_valid_aeronet_filename]
        self.file_format = "ASCIIAERONET/2"

    @istest
    def test_create_data_object_from_multiple_files(self):
        self.product().create_data_object(self.filenames, self.valid_variable)


class TestASCII(ProductTests, unittest.TestCase):
    def setUp(self):
        from cis.test.integration_test_data import ascii_filename_with_no_values
        self.setup(cis_test_files["ascii"], ASCII_Hyperpoints)
        self.no_value_filename = ascii_filename_with_no_values

    @raises(IOError)
    def test_should_raise_error_when_variable_does_not_exist_in_file(self):
        """
         This product throws an IO error rather than an InvalidVariable error as the file can only have one variable
        :return:
        """
        self.product().create_data_object([self.no_value_filename], True)

    @istest
    def test_create_data_object_with_missing_values(self):
        """
         Check that missing values get masked correctly
        :return:
        """
        data = self.product().create_data_object([self.filename], True)
        assert (all(data.data.mask == [False, False, False, True, False, True, False, False]))

    @istest
    def test_create_data_object_with_valid_datetime(self):
        import datetime
        from cis.time_util import convert_datetime_to_std_time

        data = self.product().create_data_object([self.filename], True)
        assert(data.coord('time').data[3] == convert_datetime_to_std_time(datetime.datetime(2012, 8, 25, 15, 32, 0o3)))
        assert(data.coord('time').data[4] == convert_datetime_to_std_time(datetime.datetime(2012, 8, 26)))


class TestNetCDF_Gridded_xenida(ProductTests, unittest.TestCase):
    def setUp(self):
        from cis.test.integration_test_data import valid_xenida_filename, valid_xenida_variable

        self.filename = valid_xenida_filename
        self.valid_variable = valid_xenida_variable
        self.product = NetCDF_Gridded
        self.file_format = "NetCDF/Gridded"

    @nottest
    def test_variable_wildcarding(self):
        # There are no valid variables in this file.
        pass

    @nottest
    def test_write_coords(self):
        # Model data is gridded so IRIS takes care of this
        pass

    @nottest
    def test_create_coords(self):
        # Model data is gridded so IRIS takes care of this
        pass

    @nottest
    def test_create_pandas_object(self):
        # This is broken - but it's in Iris
        pass


class TestNetCDF_Gridded_xglnwa(ProductTests, unittest.TestCase):
    def setUp(self):
        from cis.test.integration_test_data import valid_1d_filename, valid_1d_variable

        self.filename = valid_1d_filename
        self.valid_variable = valid_1d_variable
        self.product = NetCDF_Gridded
        self.file_format = "NetCDF/Gridded"

    @nottest
    def test_variable_wildcarding(self):
        # Not tested
        pass

    @nottest
    def test_write_coords(self):
        # Model data is gridded so IRIS takes care of this
        pass

    @nottest
    def test_create_coords(self):
        # Model data is gridded so IRIS takes care of this
        pass

    @nottest
    def test_create_pandas_object(self):
        # This is broken - but it's in Iris
        pass


class TestNetCDF_Gridded_HadGEM(ProductTests, unittest.TestCase):
    def setUp(self):
        from cis.test.integration_test_data import valid_hadgem_filename, valid_hadgem_variable

        self.filename = valid_hadgem_filename
        self.valid_variable = valid_hadgem_variable
        self.product = NetCDF_Gridded
        self.vars = ['od550aer']
        self.file_format = "NetCDF/Gridded"

    @nottest
    def test_write_coords(self):
        # Gridded coordinates are taken care of by IRIS
        pass

    @nottest
    def test_create_pandas_object(self):
        # Cannot create pandas objects for data with > 2 dimensions
        pass




"""
module to test the hdf4 utility function of hdf_vd.py
"""
from nose.tools import istest, eq_

import cis.data_io.hdf_vd as hdf_vd
from cis.test.integration_test_data import valid_hdf_vd_file, escape_colons, skip_pyhdf


@istest
@skip_pyhdf
def test_that_can_read_all_variables():
    dict = hdf_vd.get_hdf_VD_file_variables(escape_colons(valid_hdf_vd_file))
    eq_(len(dict), 37)
    eq_(True, 'Longitude' in dict)


@istest
@skip_pyhdf
def test_that_can_get_data():
    vds_dict = hdf_vd.read(escape_colons(valid_hdf_vd_file), 'DEM_elevation')
    vds = vds_dict['DEM_elevation']
    data = hdf_vd.get_data(vds)
    eq_(37081, len(data))


@istest
@skip_pyhdf
def test_that_can_get_variable_metadata():
    vd = hdf_vd.read(escape_colons(valid_hdf_vd_file), 'DEM_elevation')['DEM_elevation']
    metadata = hdf_vd.get_metadata(vd)
    eq_(metadata._name, "DEM_elevation")
    eq_(metadata.long_name, "Digital Elevation Map")
    eq_(metadata.shape, [37081])
    eq_(metadata.units, "meters")
    eq_(metadata.factor, 1.0)
    eq_(metadata.offset, 0.0)
    eq_(metadata.missing_value, 9999)
    eq_(metadata.misc['valid_range'], [-9999, 8850])


@istest
@skip_pyhdf
def test_that_can_get_coord_metadata():
    vd = hdf_vd.read(escape_colons(valid_hdf_vd_file), 'Longitude')['Longitude']
    metadata = hdf_vd.get_metadata(vd)
    eq_(metadata._name, "Longitude")
    eq_(metadata.standard_name, "longitude")
    eq_(metadata.long_name, "Spacecraft Longitude")
    eq_(metadata.shape, [37081])
    eq_(metadata.units, "degrees")
    eq_(metadata.factor, 1.0)
    eq_(metadata.offset, 0.0)
    eq_(metadata.missing_value, None)
    eq_(metadata.misc['valid_range'], [-180.0, 180.0])



from nose.tools import istest, eq_, raises
from cis.data_io.products.caliop import Caliop_L2
from cis.exceptions import ClassNotFoundError
from cis.data_io.products.AProduct import __get_class
from cis.parse import parse_args

# Note that the below is only used as a filename to test the product matching routines - there is no need for the actual
#  file to be present
example_caliop_l2_filename = "CAL_LID_L2_05kmAPro-Prov-V3-01.2009-12-31T23-36-08ZN.hdf"


@istest
def can_overide_default_product():
    from cis.data_io.products.gridded_NetCDF import NetCDF_Gridded
    eq_(__get_class(example_caliop_l2_filename), Caliop_L2)
    eq_(__get_class(example_caliop_l2_filename, "NetCDF_Gridded"), NetCDF_Gridded)


@istest
def should_raise_error_with_unknown_product_specified():
    try:
        parse_args(["plot", "var:" + example_caliop_l2_filename + "::::unknownproduct"])
        assert False
    except SystemExit as e:
        if e.code != 2:
            raise e


@istest
@raises(ClassNotFoundError)
def files_which_match_some_of_regexp_but_not_the_end_shouldnt_match():
    """
    Because we're using re.match the regular expression has to match the start - but not necassarily the end. This
     test ensures that we have taken care of that as often the extension is the most important part.
    """
    import cis.plugin as plugin
    from cis.data_io.products.AProduct import AProduct
    # We have to patch all of the plugin classes because get_file_type_error gets called and this file doesn't exist
    #  we are only testing the wildcard matching logic.
    product_classes = plugin.find_plugin_classes(AProduct, 'cis.data_io.products')
    for p in product_classes:
        p.get_file_type_error = lambda self, f: None
    _ = __get_class(example_caliop_l2_filename+".ext")



import unittest

from hamcrest import *

from cis.collocation import data_index
from cis.data_io.hyperpoint import HyperPoint
from cis.collocation.col_implementations import make_coord_map, BinnedCubeCellOnlyConstraint
from cis.test.util.mock import *
from cis.time_util import convert_datetime_to_std_time


class TestGridCellBinIndex(unittest.TestCase):

    def test_GIVEN_single_point_in_cube_WHEN_iterate_THEN_return_point_in_middle(self):

        sample_cube = make_square_5x3_2d_cube_with_time(offset=0, time_offset=0)
        data_point = make_dummy_ungridded_data_single_point(0.5, 0.5, 1.2, time=datetime.datetime(1984, 8, 28, 0, 0))
        coord_map = make_coord_map(sample_cube, data_point)
        coords = sample_cube.coords()
        for (hpi, ci, shi) in coord_map:
            coord = coords[ci]
            if coord.ndim > 1:
                raise NotImplementedError("Co-location of data onto a cube with a coordinate of dimension greater"
                                          " than one is not supported (coordinate %s)", coord.name())
            # Ensure that bounds exist.
            if not coord.has_bounds():
                coord.guess_bounds()

        constraint = BinnedCubeCellOnlyConstraint()
        data_index.create_indexes(constraint, coords, data_point.get_non_masked_points(), coord_map)
        iterator = constraint.get_iterator(False, coord_map, coords, data_point.get_non_masked_points(), None,
                                           sample_cube, None)

        final_points_index = [(out_index, hp, points) for out_index, hp, points in iterator]
        assert_that(len(final_points_index), is_(1), "There is one mapping from sample_cube to the final grid")
        assert_that(final_points_index[0][0], is_((2, 1, 1)), "The points should map to index")
        assert_that(final_points_index[0][1], is_(HyperPoint(lat=0, lon=0, t=datetime.datetime(1984, 8, 28))),
                    "The points should map to index")
        assert_that(final_points_index[0][2].latitudes, is_([0.5]), "The points should map to index")
        assert_that(final_points_index[0][2].longitudes, is_([0.5]), "The points should map to index")
        assert_that(final_points_index[0][2].times,
                    is_([convert_datetime_to_std_time(datetime.datetime(1984, 8, 28, 0, 0))]),
                    "The points should map to index")
        assert_that(final_points_index[0][2].vals, is_([1.2]), "The points should map to index")

    def test_GIVEN_single_masked_point_in_cube_WHEN_iterate_THEN_return_no_points(self):

        sample_cube = make_square_5x3_2d_cube_with_time(offset=0, time_offset=0)
        data_point = make_dummy_ungridded_data_single_point(0.5, 0.5, 1.2, time=datetime.datetime(1984, 8, 28, 0, 0),
                                                            mask=True)
        coord_map = make_coord_map(sample_cube, data_point)
        coords = sample_cube.coords()
        for (hpi, ci, shi) in coord_map:
            coord = coords[ci]
            if coord.ndim > 1:
                raise NotImplementedError("Co-location of data onto a cube with a coordinate of dimension greater"
                                          " than one is not supported (coordinate %s)", coord.name())
            # Ensure that bounds exist.
            if not coord.has_bounds():
                coord.guess_bounds()

        constraint = BinnedCubeCellOnlyConstraint()
        data_index.create_indexes(constraint, coords, data_point.get_non_masked_points(), coord_map)
        iterator = constraint.get_iterator(False, coord_map, coords, data_point.get_non_masked_points(), None,
                                           sample_cube, None)

        final_points_index = [(out_index, hp, points) for out_index, hp, points in iterator]
        assert_that(len(final_points_index), is_(0), "Masked points should not be iterated over")



"""
Module that contains various strings that are used in tests
"""
from collections import namedtuple
from datetime import datetime
from unittest import skipIf
import os

try:
    import pyhdf
except ImportError:
    # Disable all these tests if pandas is not installed.
    pyhdf = None

skip_pyhdf = skipIf(pyhdf is None, 'Test(s) require "pandas", which is not available.')

# A dictionary of test file tuples indexed by characteristic name
cis_test_files = {}

# values are None for not applicable
TestFileTestData = namedtuple('TestFile',
                              ["master_filename",
                               "file_format",
                               "product_name",
                               "start_datetime",
                               "end_datetime",
                               "lat_min",
                               "lat_max",
                               "lon_min",
                               "lon_max",
                               "valid_vars_count",
                               "all_variable_names",
                               "data_variable_name",
                               "data_variable_standard_name",
                               "data_variable_properties"])

data_directory = os.environ.get("CIS_DATA_HOME", os.path.dirname(__file__))


def escape_colons(filename):
    if os.name == 'nt':
        filename = filename.replace(':', '\:')
    return filename


def make_pathname(filename):
    path = os.path.join(data_directory, filename)
    return path


valid_hdf_vd_file = make_pathname("2008045004519_09563_CS_2C-PRECIP-COLUMN_GRANULE_P_R04_E02.hdf")
valid_hdf_sd_file = make_pathname("MOD04_L2.A2010001.2255.005.2010005215814.hdf")

valid_modis_l2_filename = make_pathname("MOD04_L2.A2010001.2255.005.2010005215814.hdf")
valid_modis_l2_variable = 'Cloud_Fraction_Ocean'

valid_modis_l2_variables = \
    ['Deep_Blue_Angstrom_Exponent_Land',
     'Cloud_Fraction_Ocean',
     'Corrected_Optical_Depth_Land_wav2p1',
     'Mass_Concentration_Land',
     'Solar_Zenith',
     'Latitude',
     'Sensor_Azimuth',
     'Optical_Depth_Ratio_Small_Land_And_Ocean',
     'Sensor_Zenith',
     'Scan_Start_Time',
     'Image_Optical_Depth_Land_And_Ocean',
     'Cloud_Fraction_Land',
     'Number_Pixels_Used_Ocean',
     'Longitude',
     'Aerosol_Type_Land',
     'Cloud_Mask_QA',
     'Optical_Depth_Ratio_Small_Land',
     'Scattering_Angle',
     'Solar_Azimuth',
     'Angstrom_Exponent_Land',
     'Deep_Blue_Aerosol_Optical_Depth_550_Land',
     'Fitting_Error_Land',
     'Optical_Depth_Land_And_Ocean']

cis_test_files["modis_L2"] = TestFileTestData(
    master_filename=valid_modis_l2_filename,
    file_format="HDF4/ModisL2",
    product_name="MODIS_L2",
    start_datetime=datetime(2010, 1, 1, 22, 55, 8),
    end_datetime=datetime(2010, 1, 1, 23, 00, 6),
    lat_min=-80.405128,
    lat_max=-56.644821,
    lon_min=-179.99913,
    lon_max=179.99573,
    valid_vars_count=len(valid_modis_l2_variables),
    all_variable_names=valid_modis_l2_variables,
    data_variable_name=valid_modis_l2_variable,
    data_variable_standard_name=valid_modis_l2_variables,
    data_variable_properties={}
)

modis_myd06_l2_filenames = [make_pathname("MYD06_L2.A2011100.1720.051.2011102130126.hdf"),
                            make_pathname("MYD06_L2.A2011100.1725.051.2011102123922.hdf")]
modis_myd06_l2_variable = "Surface_Temperature"

valid_modis_l3_filename = make_pathname("MOD08_E3.A2010009.005.2010026072315.hdf")
valid_modis_l3_filename2 = make_pathname("MOD08_E3.A2010017.005.2010029140340.hdf")
valid_modis_l3_variable = "Solar_Zenith_Std_Deviation_Mean"
cis_test_files["modis_L3"] = TestFileTestData(
    master_filename=valid_modis_l3_filename,
    file_format="HDF4/ModisL3",
    product_name="MODIS_L3",
    start_datetime=datetime(2010, 1, 13, 0, 0, 0),
    end_datetime=datetime(2010, 1, 13, 0, 0, 0),
    lat_min=-89.5,
    lat_max=89.5,
    lon_min=-179.5,
    lon_max=179.5,
    valid_vars_count=700,
    all_variable_names=None,
    data_variable_name="Solar_Zenith_Mean_Mean",
    data_variable_standard_name="Solar_Zenith_Mean_Mean",
    data_variable_properties={}
)

cloudsat_file_format = "HDF4/CloudSat"
valid_cloudsat_RVOD_file = make_pathname("2007180125457_06221_CS_2B-CWC-RVOD_GRANULE_P_R04_E02.hdf")
valid_cloudsat_RVOD_sdata_variable = "RVOD_liq_water_content"
valid_cloudsat_RVOD_vdata_variable = "RVOD_ice_water_path"
valid_cloudsat_RVOD_file_format = cloudsat_file_format

cis_test_files["CloudsatRVODsdata"] = TestFileTestData(
    master_filename=valid_cloudsat_RVOD_file,
    file_format=cloudsat_file_format,
    product_name="CloudSat",
    start_datetime=datetime(2007, 6, 29, 12, 55, 3),
    end_datetime=datetime(2007, 6, 29, 14, 33, 56),
    lat_min=-81.84876,
    lat_max=81.849,
    lon_min=-180,
    lon_max=180,
    valid_vars_count=69,
    all_variable_names=None,
    data_variable_name=valid_cloudsat_RVOD_sdata_variable,
    data_variable_standard_name=valid_cloudsat_RVOD_sdata_variable,
    data_variable_properties={}
)

valid_cloudsat_PRECIP_file = make_pathname("2008045004519_09563_CS_2C-PRECIP-COLUMN_GRANULE_P_R04_E02.hdf")
valid_cloudsat_PRECIP_variable = "Precip_rate"
valid_cloudsat_PRECIP_file_format = cloudsat_file_format

valid_aerosol_cci_filename = make_pathname("20080612093821-ESACCI-L2P_AEROSOL-ALL-AATSR_ENVISAT-ORAC_32855-fv02.02.nc")
valid_aerosol_cci_variable = "AOD550"
valid_aerosol_cci_variable_2 = "AOD870"

cis_test_files["Aerosol_CCI"] = TestFileTestData(
    master_filename=valid_aerosol_cci_filename,
    file_format="NetCDF/Aerosol_CCI_L2",
    product_name="Aerosol_CCI_L2",
    start_datetime=datetime(2008, 6, 12, 10, 1, 23),
    end_datetime=datetime(2008, 6, 12, 10, 43, 37),
    lat_min=-44.91,
    lat_max=83.80,
    lon_min=-22.25,
    lon_max=160.48,
    valid_vars_count=32,
    all_variable_names=[valid_aerosol_cci_variable, valid_aerosol_cci_variable_2],
    data_variable_name=valid_aerosol_cci_variable,
    data_variable_standard_name='atmosphere_optical_thickness_due_to_aerosol',
    data_variable_properties={}
)

valid_cloud_cci_filename = make_pathname("20080620072500-ESACCI-L2_CLOUD-CLD_PRODUCTS-MODIS-AQUA-fv1.0.nc")
valid_cloud_cci_variable = "stemp"
valid_cloud_cci_8_bit_variable = "illum"

cis_test_files["Cloud_CCI"] = TestFileTestData(
    master_filename=valid_cloud_cci_filename,
    file_format="NetCDF/Cloud_CCI_L2",
    product_name="Cloud_CCI_L2",
    start_datetime=datetime(2007, 6, 29, 12, 55, 3),
    end_datetime=datetime(2007, 6, 29, 14, 33, 56),
    lat_min=-81.84876,
    lat_max=81.849,
    lon_min=-180,
    lon_max=180,
    valid_vars_count=30,
    all_variable_names=None,
    data_variable_name=valid_cloud_cci_variable,
    data_variable_standard_name=valid_cloud_cci_variable,
    data_variable_properties={}
)

valid_hadgem_filename = make_pathname('od550aer.nc')
valid_hadgem_variable = 'od550aer'

valid_aeronet_filename = make_pathname("920801_091128_Agoufou_small.lev20")
another_valid_aeronet_filename = make_pathname("920801_091128_Abracos_Hill_small.lev20")
valid_aeronet_variable = "AOT_440"
all_aeronet_variables = [
    "AOT_551",
    "TripletVar_532",
    "AOT_532",
    "AOT_667",
    "AOT_531",
    "TripletVar_412",
    "440-675Angstrom(Polar)",
    "AOT_1640",
    "WaterError",
    "AOT_380",
    "500-870Angstrom",
    "Date(dd-mm-yy)",
    "AOT_412",
    "AOT_555",
    "AOT_490",
    "TripletVar_1640",
    "Solar_Zenith_Angle",
    "340-440Angstrom",
    "TripletVar_340",
    "AOT_675",
    "TripletVar_531",
    "TripletVar_380",
    "TripletVar_440",
    "TripletVar_500",
    "380-500Angstrom",
    "AOT_443",
    "AOT_440",
    "TripletVar_667",
    "440-675Angstrom",
    "Time(hh:mm:ss)",
    "TripletVar_675",
    "Julian_Day",
    "TripletVar_1020",
    "TripletVar_870",
    "TripletVar_443",
    "Water(cm)",
    "AOT_500",
    "AOT_1020",
    "440-870Angstrom",
    "AOT_870",
    "TripletVar_490",
    "Last_Processing_Date",
    "TripletVar_551",
    "AOT_340",
    "TripletVar_555"
]

cis_test_files["aeronet"] = TestFileTestData(
    master_filename=valid_aeronet_filename,
    file_format="Aeronet",
    product_name="Aeronet",
    start_datetime=datetime(2003, 9, 25, 6, 47, 9),
    end_datetime=datetime(2003, 12, 31, 9, 24, 4),
    lat_min=15.345,
    lat_max=15.345,
    lon_min=-1.479,
    lon_max=-1.479,
    valid_vars_count=len(all_aeronet_variables),
    all_variable_names=all_aeronet_variables,
    data_variable_name="AOT_440",
    data_variable_standard_name="AOT_440",
    data_variable_properties={}
)

# 6 hourly average of the above data
aggregated_aeronet_filename = make_pathname("Abracos_Hill_AOT_6_hourly.nc")
aggregated_aeronet_variables = ["AOT_440",
                                "AOT_440_std_dev"]

valid_caliop_l2_filename = make_pathname("CAL_LID_L2_05kmAPro-Prov-V3-01.2009-12-31T23-36-08ZN.hdf")
valid_caliop_l2_variable = "Pressure"
valid_caliop_l2_variables = ['Aerosol_Layer_Fraction',
                             'Aerosol_Multiple_Scattering_Profile_1064',
                             'Aerosol_Multiple_Scattering_Profile_532',
                             'Backscatter_Coefficient_1064',
                             'Backscatter_Coefficient_Uncertainty_1064',
                             'Cloud_Layer_Fraction',
                             'Extinction_Coefficient_1064',
                             'Extinction_Coefficient_532',
                             'Extinction_Coefficient_Uncertainty_1064',
                             'Extinction_Coefficient_Uncertainty_532',
                             'Molecular_Number_Density',
                             'Particulate_Depolarization_Ratio_Profile_532',
                             'Particulate_Depolarization_Ratio_Uncertainty_532',
                             'Perpendicular_Backscatter_Coefficient_532',
                             'Perpendicular_Backscatter_Coefficient_Uncertainty_532',
                             'Pressure',
                             'Relative_Humidity',
                             'Samples_Averaged',
                             'Temperature',
                             'Total_Backscatter_Coefficient_Uncertainty_532',
                             'Total_Backscatter_Coefficient_532']
cis_test_files["caliop_L2"] = TestFileTestData(
    master_filename=valid_caliop_l2_filename,
    file_format="HDF4/CaliopL2",
    product_name="Caliop_L2",
    start_datetime=datetime(2009, 12, 31, 23, 36, 12),
    end_datetime=datetime(2010, 1, 1, 0, 22, 25),
    lat_min=-61.37,
    lat_max=81.84,
    lon_min=-179.97,
    lon_max=179.93,
    valid_vars_count=len(valid_caliop_l2_variables),
    all_variable_names=valid_caliop_l2_variables,
    data_variable_name="Cloud_Layer_Fraction",
    data_variable_standard_name="Cloud_Layer_Fraction",
    data_variable_properties={}
)

valid_caliop_l1_filename = make_pathname("CAL_LID_L1-ValStage1-V3-01.2009-12-31T23-36-08ZN.hdf")
valid_caliop_l1_variable = "Attenuated_Backscatter_1064"
valid_caliop_l1_variables = ['Perpendicular_Attenuated_Backscatter_532',
                             'Attenuated_Backscatter_1064',
                             'Total_Attenuated_Backscatter_532']
cis_test_files["caliop_L1"] = TestFileTestData(
    master_filename=valid_caliop_l1_filename,
    file_format="HDF4/CaliopL1",
    product_name="Caliop_L1",
    start_datetime=datetime(2009, 12, 31, 23, 36, 12),
    end_datetime=datetime(2010, 1, 1, 0, 22, 25),
    lat_min=-61.37,
    lat_max=81.84,
    lon_min=-179.97,
    lon_max=179.93,
    valid_vars_count=len(valid_caliop_l1_variables),
    all_variable_names=valid_caliop_l1_variables,
    data_variable_name='Perpendicular_Attenuated_Backscatter_532',
    data_variable_standard_name='Perpendicular_Attenuated_Backscatter_532',
    data_variable_properties={}
)

valid_echamham_filename = make_pathname("ECHAMHAM_AOT550_670.nc")
valid_echamham_variable_1 = "TAU_2D_550nm"
valid_echamham_variable_2 = "TAU_2D_670nm"

valid_xenida_filename = make_pathname("xenida.pah9440.nc")
valid_xenida_variable = "specific_humidity"

valid_1d_filename = make_pathname("xglnwa.pm.k8dec-k9nov.vprof.tm.nc")
valid_1d_variable = "q"
valid_2d_filename = make_pathname("xglnwa.pm.k8dec-k9nov.col.tm.nc")

# TODO These two test files are the same but appear to be testing different things...
valid_echamham_geopotential_filename = make_pathname('ECHAM6.1-HAM2.2_GLOFIR1_2008_mmrbc_zgeopotm.nc')
valid_echamham_geopotential_variable = 'mmrbc'

valid_echamham_geopotential_height_filename = make_pathname('ECHAM6.1-HAM2.2_GLOFIR1_2008_mmrbc_zgeopotm.nc')
valid_echamham_geopotential_height_variable = 'mmrbc'

cis_test_files["2D_GRIDDED"] = TestFileTestData(
    master_filename=valid_2d_filename,
    file_format="NetCDF/Gridded",
    product_name="defaultNetCDF",
    start_datetime=None,
    end_datetime=None,
    lat_min=None,
    lat_max=None,
    lon_min=None,
    lon_max=None,
    valid_vars_count=None,
    all_variable_names=[],
    data_variable_name="SST_1",
    data_variable_standard_name="SST_1",
    data_variable_properties=None
)

valid_cis_col_file = make_pathname("cis-col-latlon-renamed.nc")
valid_cis_col_variable = 'AOT_440'

valid_cis_gridded_output_filename = make_pathname("subset-gridded-out.nc")
valid_cis_gridded_output_variable = 'TAU_2D_550nm'

valid_cis_ungridded_output_filename = make_pathname('cis-subset-ungridded-out.nc')
valid_cis_ungridded_output_variable = 'AOD550'
valid_cis_ungridded_output_variables = ['pixel_number', 'latitude', 'longitude', 'time', 'AOD550', 'AOD870']

cis_test_files["CIS_Ungridded"] = TestFileTestData(
    master_filename=valid_cis_ungridded_output_filename,
    file_format="NetCDF/CIS",
    product_name="cis",
    start_datetime=datetime(2008, 6, 12, 10, 18, 17),
    end_datetime=datetime(2008, 6, 12, 10, 20, 10),
    lat_min=39.07,
    lat_max=45.0,
    lon_min=0,
    lon_max=6.0,
    valid_vars_count=len(valid_cis_ungridded_output_variables),
    all_variable_names=valid_cis_ungridded_output_variables,
    data_variable_name=valid_cis_ungridded_output_variable,
    data_variable_standard_name='atmosphere_optical_thickness_due_to_aerosol',
    data_variable_properties={}
)

valid_ascii_filename = make_pathname("my_dummy_points_with_values.txt")
ascii_filename_with_no_values = make_pathname("my_dummy_points.txt")
cis_test_files["ascii"] = TestFileTestData(
    master_filename=valid_ascii_filename,
    file_format="ASCII/ASCIIHyperpoints",
    product_name="ASCII_Hyperpoints",
    start_datetime=datetime(2012, 8, 22, 15, 32, 3),
    end_datetime=datetime(2012, 8, 29, 15, 32, 3),
    lat_min=0,
    lat_max=7,
    lon_min=-10,
    lon_max=0,
    valid_vars_count=1,
    all_variable_names=['value'],
    data_variable_name='value',
    data_variable_standard_name='value',
    data_variable_properties={}
)

netcdf_cf_compliant_ranges_filename = make_pathname("netcdf_file_with_valid_ranges.nc")
netcdf_cf_compliant_ranges_vars = ["LON_502", "SO4_1950", "CL_1944"]

valid_NCAR_NetCDF_RAF_filename = make_pathname("RF04.20090114.192600_035100.PNI.nc")
valid_NCAR_NetCDF_RAF_variable = 'ATX'

cis_test_files["NCAR_NetCDF_RAF"] = TestFileTestData(
    master_filename=make_pathname("RF04.20090114.192600_035100.PNI.nc"),
    file_format="NetCDF/NCAR-RAF/nimbus/1.3",
    product_name="NCAR_NetCDF_RAF",
    start_datetime=datetime(2009, 1, 14, 19, 26, 00),
    end_datetime=datetime(2009, 1, 15, 3, 51, 30),
    lat_min=29.4966,
    lat_max=31.0048,
    lon_min=20,
    lon_max=62,
    #    alt_min=20.5,
    #    alt_max=3678.5,
    valid_vars_count=121,
    all_variable_names=None,
    data_variable_name='ATX',
    data_variable_standard_name="ATX",
    data_variable_properties={
        "units": "ng/kg",
        "missing_value": -9999}
)

# this data is not entirely correct because there is an extra 1 length dimension in it
valid_GASSP_not_entirely_correct_filename = make_pathname("ACE1_aero_an.nc")
valid_GASSP_not_entirely_correct_variable = 'TIM_627'

cis_test_files["GASSP_aux_coord"] = TestFileTestData(
    master_filename=make_pathname("ACE1_NSD_RF06_11081995163707_11081995210628.nc"),
    file_format="NetCDF/GASSP/2.0",
    product_name="NCAR_NetCDF_RAF",
    start_datetime=datetime(1995, 11, 8, 16, 37, 0o7),
    end_datetime=datetime(1995, 11, 8, 21, 6, 28),
    lat_min=4.85591,
    lat_max=21.3232,
    lon_min=-157.946,
    lon_max=-155.102,
    #    alt_min=20.5,
    #    alt_max=3678.5,
    all_variable_names=["NUMDIST_DMA_OPC", "AREADIST_DMA_OPC", "VOLDIST_DMA_OPC",],
    valid_vars_count=11,
    data_variable_name='NUMDIST_DMA_OPC',
    data_variable_standard_name='NUMDIST_DMA_OPC',
    data_variable_properties={
        "units": "#/cm3",
        "missing_value": -32767}
)

valid_GASSP_aeroplane_filename = make_pathname("SP2_mrg60_NP3_20060927_R1.ict.nc")
valid_GASSP_aeroplane_vars = ['BC_ng_kg', "BC_ng_m3", "UTC_mid", "GpsAlt", "GpsLat", "GpsLon"]
valid_GASSP_aeroplane_variable = valid_GASSP_aeroplane_vars[0]

cis_test_files["GASSP_aeroplane"] = TestFileTestData(
    master_filename=valid_GASSP_aeroplane_filename,
    file_format="NetCDF/GASSP/1.0",
    product_name="NCAR_NetCDF_RAF",
    start_datetime=datetime(2006, 9, 27, 18, 43, 30),
    end_datetime=datetime(2006, 9, 27, 23, 53, 30),
    lat_min=29.4966,
    lat_max=31.0048,
    lon_min=-96.0271,
    lon_max=-93.4528,
    #    alt_min=20.5,
    #    alt_max=3678.5,
    all_variable_names=valid_GASSP_aeroplane_vars,
    valid_vars_count=len(valid_GASSP_aeroplane_vars),
    data_variable_name=valid_GASSP_aeroplane_variable,
    data_variable_standard_name=valid_GASSP_aeroplane_variable,
    data_variable_properties={
        "units": "ng/kg",
        "missing_value": -9999}
)

cis_test_files["GASSP_ship"] = TestFileTestData(
    master_filename=make_pathname("AMSpmel_RHB_20081020_R0.ict.nc"),
    file_format="NetCDF/GASSP/1.0",
    product_name="NCAR_NetCDF_RAF",
    start_datetime=datetime(2008, 10, 23, 00, 00, 00),
    end_datetime=datetime(2008, 10, 27, 00, 00, 00),
    lat_min=29.4966,
    lat_max=31.0048,
    lon_min=-96.0271,
    lon_max=-93.4528,
    #    alt_min=20.5,
    #    alt_max=3678.5,
    all_variable_names=['NRno3', "Start_UTC", "LAT_Deg", "LONGT_Deg", "NRso4", "NRpom", "NRnh4"],
    valid_vars_count=7,
    data_variable_name='NRno3',
    data_variable_standard_name='NRno3',
    data_variable_properties=None
)
valid_GASSP_ship_vars = cis_test_files["GASSP_ship"].all_variable_names
valid_GASSP_ship_filename = cis_test_files["GASSP_ship"].master_filename

valid_GASSP_attribute = 'Time_Coordinate'

cis_test_files["GASSP_station"] = TestFileTestData(
    master_filename=make_pathname("TrinidadHead.US6005G.20110101.20140222.aerosol_number_concentration.aerosol.1y.1h."
                                  "US06L_TSI_3760_THD.US06L_cpc_ref.nas.txt.nc"),
    file_format="NetCDF/GASSP/1.0",
    product_name="NCAR_NetCDF_RAF",
    start_datetime=datetime(2008, 10, 23, 00, 00, 00),
    end_datetime=datetime(2008, 10, 27, 00, 00, 00),
    lat_min=29.4966,
    lat_max=31.0048,
    lon_min=-96.0271,
    lon_max=-93.4528,
    #    alt_min=20.5,
    #    alt_max=3678.5,
    all_variable_names=['aerosol_number_concentration',
                        "start_time",
                        "end_time",
                        "aerosol_number_concentration_statistics_percentile_84.13",
                        "aerosol_number_concentration_statistics_percentile_15.87",
                        "numflag"],
    valid_vars_count=6,
    data_variable_name='aerosol_number_concentration',
    data_variable_standard_name='aerosol_number_concentration',
    data_variable_properties=None
)

valid_GASSP_station_filename = cis_test_files["GASSP_station"].master_filename
valid_GASSP_station_vars = cis_test_files["GASSP_station"].all_variable_names

valid_GASSP_station_files_with_different_timestamps = [make_pathname('CCN_mrg60_NP3_20080329_R6.ict.nc'),
                                                       make_pathname('CCN_mrg60_NP3_20080401_R6.ict.nc')]
valid_GASSP_station_var_with_different_timestamps = 'Number_Concentration'

valid_GASSP_aircraft_files_with_different_timestamps = [make_pathname('ACE1_CCN_RF11_11171995224801_11181995073133.nc'),
                                                        make_pathname('ACE1_CCN_RF12_11181995222509_11191995072145.nc')]
valid_GASSP_aircraft_var_with_different_timestamps = 'PRESSURE_ALTITUDE'

valid_zonal_time_mean_CMIP5_filename = \
    make_pathname('rsutcs_Amon_HadGEM2-A_sstClim_r1i1p1_185912-188911.CMIP5.tm.zm.nc')
valid_zonal_time_mean_CMIP5_variable = "rsutcs"

valid_hybrid_pressure_filename = make_pathname('hybrid-pressure.nc')
valid_hybrid_pressure_variable = 'mass_fraction_of_black_carbon_dry_aerosol_in_air'

valid_hybrid_height_flat_filename = make_pathname('hybrid-height.nc')
valid_hybrid_height_flat_variable = 'mass_fraction_of_black_carbon_dry_aerosol_in_air'

valid_hybrid_height_filename = make_pathname('aerocom.CAM4-Oslo.A2.CTRL.monthly.mmrbc.9999_fixed.nc')
valid_hybrid_height_variable = 'mmrbc'

# Has netCDF4 hierarchical groups
valid_netcdf_groups_file = make_pathname('W_XX-EUMETSAT-Darmstadt_SOUNDING_and_SATELLITE_METOPB_and_'
                                         'GOME_C_EUMP_20140405235655_08032_eps_t_pmap_l2.nc')
valid_netcdf_groups_variable = 'AVHRR/Ch4CentralWavenumber'
valid_nested_groups_file = make_pathname('nested_groups.nc')
valid_nested_groups_variable = 'group1/group2/var4'

invalid_filename = "invalidfilename"
non_netcdf_file = make_pathname("notanetcdffile")
non_netcdf_file_with_netcdf_file_extension = make_pathname("notarealnetcdffile.nc")



from nose.tools import istest, nottest, raises
import numpy as np

from cis.data_io.hyperpoint_view import UngriddedHyperPointView, GriddedHyperPointView
import cis.test.util.mock as mock
import cis.data_io.gridded_data as gridded_data


class TestUngriddedHyperPointView(object):
    """
    Unit tests for UngriddedHyperPointView class
    """
    @istest
    def test_can_create_ungridded_hyper_point_view(self):
        ug = mock.make_regular_2d_ungridded_data_with_missing_values()
        all_coords = ug._coords.find_standard_coords()
        flattened_coords = [(c.data_flattened if c is not None else None) for c in all_coords]
        hpv = UngriddedHyperPointView(flattened_coords, ug.data_flattened)
        assert(len(hpv) == 15)

    @istest
    def test_can_access_point_in_ungridded_hyper_point_view(self):
        ug = mock.make_regular_2d_ungridded_data_with_missing_values()
        hpv = ug.get_all_points()
        p = hpv[10]
        assert(p.val[0] == 11.0)
        assert(p.longitude == 0.0)
        assert(p.latitude == 5.0)

    @istest
    @raises(IndexError)
    def test_accessing_with_invalid_index_raises_error(self):
        ug = mock.make_regular_2d_ungridded_data_with_missing_values()
        hpv = ug.get_all_points()
        p = hpv[16]

    @istest
    def test_can_iterate_over_all_points(self):
        ug = mock.make_regular_2d_ungridded_data()
        hpv = ug.get_all_points()
        count = 0
        vals = []
        for p in hpv:
            count += 1
            vals.append(p.val[0])
        assert(count == 15)
        assert(vals == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15])

    @istest
    def test_can_iterate_over_non_masked_points(self):
        ug = mock.make_regular_2d_ungridded_data_with_missing_values()
        hpv = ug.get_non_masked_points()
        count = 0
        vals = []
        for p in hpv:
            count += 1
            vals.append(p.val[0])
        assert(count == 12)
        assert(vals == [1, 2, 3, 4, 6, 7, 8, 10, 11, 12, 14, 15])

    @istest
    def test_can_enumerate_non_masked_points(self):
        ug = mock.make_regular_2d_ungridded_data_with_missing_values()
        hpv = ug.get_non_masked_points()
        count = 0
        indices = []
        vals = []
        for i, p in hpv.enumerate_non_masked_points():
            count += 1
            indices.append(i)
            vals.append(p.val[0])
        assert(count == 12)
        assert(indices == [0, 1, 2, 3, 5, 6, 7, 9, 10, 11, 13, 14])
        assert(vals == [1, 2, 3, 4, 6, 7, 8, 10, 11, 12, 14, 15])

    @istest
    def test_can_set_a_hyperpoint(self):
        ug = mock.make_regular_2d_ungridded_data_with_missing_values()
        hpv = ug.get_all_points()
        p = hpv[6]
        p_new = p.modified(lon=123, val=99)
        hpv[6] = p_new
        p = hpv[6]
        assert(p.latitude == 0)
        assert(p.longitude == 123)
        assert(p.val[0] == 99)
        # Check an unmodified point.
        p = hpv[5]
        assert(p.latitude == -5)
        assert(p.longitude == 5)
        assert(p.val[0] == 6)

    @istest
    def test_can_modify_a_hyperpoint_value(self):
        ug = mock.make_regular_2d_ungridded_data_with_missing_values()
        hpv = ug.get_all_points()
        hpv[6] = 99
        p = hpv[6]
        assert(p.latitude == 0)
        assert(p.longitude == -5)
        assert(p.val[0] == 99)
        # Check an unmodified point.
        p = hpv[5]
        assert(p.latitude == -5)
        assert(p.longitude == 5)
        assert(p.val[0] == 6)

    @istest
    def test_can_set_longitude_wrap_at_180(self):
        ug = mock.make_regular_2d_ungridded_data(lat_dim_length=5, lat_min=-90, lat_max=90,
                                                 lon_dim_length=9, lon_min=0, lon_max=360)
        hpv = ug.get_non_masked_points()
        hpv.set_longitude_range(-180.0)
        assert(np.min(hpv.longitude) >= -180.0)
        assert(np.max(hpv.longitude) < 180.0)
        assert(hpv[0].longitude == 0.0)
        assert(hpv[36].longitude == 0.0)
        assert(hpv[4].longitude == -180.0)
        assert(hpv[40].longitude == -180.0)
        assert(hpv[7].longitude == -45.0)
        assert(hpv[43].longitude == -45.0)
        assert(hpv[8].longitude == 0.0)
        assert(hpv[44].longitude == 0.0)

    @istest
    def test_can_set_longitude_wrap_at_360(self):
        ug = mock.make_regular_2d_ungridded_data(lat_dim_length=5, lat_min=-90, lat_max=90,
                                                 lon_dim_length=9, lon_min=-180, lon_max=180)
        hpv = ug.get_non_masked_points()
        hpv.set_longitude_range(0.0)
        assert(np.min(hpv.longitude) >= 0.0)
        assert(np.max(hpv.longitude) < 360.0)
        assert(hpv[0].longitude == 180.0)
        assert(hpv[36].longitude == 180.0)
        assert(hpv[1].longitude == 225.0)
        assert(hpv[37].longitude == 225.0)
        assert(hpv[4].longitude == 0.0)
        assert(hpv[40].longitude == 0.0)
        assert(hpv[8].longitude == 180.0)
        assert(hpv[44].longitude == 180.0)


class TestGriddedHyperPointView(object):
    """
    Unit tests for GriddedHyperPointView class
    """
    @istest
    def test_can_create_gridded_hyper_point_view(self):
        gd = gridded_data.make_from_cube(mock.make_5x3_lon_lat_2d_cube_with_missing_data())
        all_coords = [((c[0].points, c[1]) if c is not None else None) for c in gd.find_standard_coords()]
        hpv = GriddedHyperPointView(all_coords, gd.data)
        assert(len(hpv) == 15)

    @istest
    def test_can_access_point_in_gridded_hyper_point_view_(self):
        gd = gridded_data.make_from_cube(mock.make_5x3_lon_lat_2d_cube_with_missing_data())
        hpv = gd.get_all_points()
        p = hpv[3, 1]
        assert(p.val[0] == 11.0)
        assert(p.longitude == 5.0)
        assert(p.latitude == 0.0)

    @istest
    @raises(IndexError)
    def test_accessing_with_invalid_index_raises_error(self):
        gd = gridded_data.make_from_cube(mock.make_5x3_lon_lat_2d_cube_with_missing_data())
        hpv = gd.get_all_points()
        p = hpv[2, 3]

    @istest
    @raises(IndexError)
    def test_accessing_with_invalid_flat_index_raises_error(self):
        gd = gridded_data.make_from_cube(mock.make_5x3_lon_lat_2d_cube_with_missing_data())
        hpv = gd.get_all_points()
        p = hpv[16]

    @istest
    def test_can_iterate_over_all_points(self):
        gd = gridded_data.make_from_cube(mock.make_square_5x3_2d_cube())
        hpv = gd.get_all_points()
        count = 0
        vals = []
        for p in hpv:
            count += 1
            vals.append(p.val[0])
        assert(count == 15)
        assert(vals == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15])

    @istest
    def test_can_iterate_over_non_masked_points(self):
        gd = gridded_data.make_from_cube(mock.make_5x3_lon_lat_2d_cube_with_missing_data())
        hpv = gd.get_non_masked_points()
        count = 0
        vals = []
        for p in hpv:
            count += 1
            vals.append(p.val[0])
        assert(count == 12)
        assert(vals == [1, 2, 3, 4, 6, 7, 8, 10, 11, 12, 14, 15])

    @istest
    def test_can_enumerate_non_masked_points(self):
        gd = gridded_data.make_from_cube(mock.make_5x3_lon_lat_2d_cube_with_missing_data())
        hpv = gd.get_non_masked_points()
        count = 0
        indices = []
        vals = []
        for i, p in hpv.enumerate_non_masked_points():
            count += 1
            indices.append(i)
            vals.append(p.val[0])
        assert(count == 12)
        assert(indices == [0, 1, 2, 3, 5, 6, 7, 9, 10, 11, 13, 14])
        assert(vals == [1, 2, 3, 4, 6, 7, 8, 10, 11, 12, 14, 15])

    @istest
    def test_can_set_a_hyperpoint_via_flat_index(self):
        gd = gridded_data.make_from_cube(mock.make_5x3_lon_lat_2d_cube_with_missing_data())
        hpv = gd.get_non_masked_points()
        p = hpv[6]
        p_new = p.modified(lon=123, val=99)
        hpv[6] = p_new
        p = hpv[6]
        assert(p.latitude == -5)
        # Note: the coordinates cannot be changed for GriddedHyperPointView.
        assert(p.longitude == 0)
        assert(p.val[0] == 99)
        # Check an unmodified point.
        p = hpv[5]
        assert(p.latitude == 5)
        assert(p.longitude == -5)
        assert(p.val[0] == 6)

    @istest
    def test_can_set_a_hyperpoint(self):
        gd = gridded_data.make_from_cube(mock.make_5x3_lon_lat_2d_cube_with_missing_data())
        hpv = gd.get_non_masked_points()
        p = hpv[2, 0]
        p_new = p.modified(lon=123, val=99)
        hpv[6] = p_new
        p = hpv[2, 0]
        assert(p.latitude == -5)
        # Note: the coordinates cannot be changed for GriddedHyperPointView.
        assert(p.longitude == 0)
        assert(p.val[0] == 99)
        # Check an unmodified point.
        p = hpv[1, 2]
        assert(p.latitude == 5)
        assert(p.longitude == -5)
        assert(p.val[0] == 6)

    @istest
    @raises(ValueError)
    def test_setting_new_hyperpoint_coord_can_raise_exception(self):
        gd = gridded_data.make_from_cube(mock.make_5x3_lon_lat_2d_cube_with_missing_data())
        hpv = gd.get_non_masked_points()
        hpv._verify_no_coord_change_on_setting = True
        p = hpv[6]
        p_new = p.modified(lon=123, val=99)
        hpv[6] = p_new

    @istest
    def test_can_modify_a_hyperpoint_value(self):
        gd = gridded_data.make_from_cube(mock.make_5x3_lon_lat_2d_cube_with_missing_data())
        hpv = gd.get_non_masked_points()
        hpv[2, 0] = 99
        p = hpv[2, 0]
        assert(p.latitude == -5)
        assert(p.longitude == 0)
        assert(p.val[0] == 99)
        # Check an unmodified point.
        p = hpv[1, 2]
        assert(p.latitude == 5)
        assert(p.longitude == -5)
        assert(p.val[0] == 6)

    @istest
    def test_can_modify_a_hyperpoint_value_via_flat_index(self):
        gd = gridded_data.make_from_cube(mock.make_5x3_lon_lat_2d_cube_with_missing_data())
        hpv = gd.get_non_masked_points()
        hpv[6] = 99
        p = hpv[6]
        assert(p.latitude == -5)
        assert(p.longitude == 0)
        assert(p.val[0] == 99)
        # Check an unmodified point.
        p = hpv[5]
        assert(p.latitude == 5)
        assert(p.longitude == -5)
        assert(p.val[0] == 6)

    @istest
    def test_can_set_longitude_wrap_at_180(self):
        gd = gridded_data.make_from_cube(mock.make_mock_cube(lat_dim_length=5, lon_dim_length=9))
        long_coord = gd.coord('longitude')
        long_coord.points = np.array([0., 45., 90., 135., 180., 225., 270., 315., 360.])
        long_coord.bounds = None
        long_coord.guess_bounds()
        hpv = gd.get_non_masked_points()
        hpv.set_longitude_range(-180.0)
        assert(np.min(hpv.longitudes) >= -180.0)
        assert(np.max(hpv.longitudes) < 180.0)
        assert(hpv[0, 0].longitude == 0.0)
        assert(hpv[4, 0].longitude == 0.0)
        assert(hpv[0, 4].longitude == -180.0)
        assert(hpv[4, 4].longitude == -180.0)
        assert(hpv[0, 7].longitude == -45.0)
        assert(hpv[4, 7].longitude == -45.0)
        assert(hpv[0, 8].longitude == 0.0)
        assert(hpv[4, 8].longitude == 0.0)

    @istest
    def test_can_set_longitude_wrap_at_360(self):
        gd = gridded_data.make_from_cube(mock.make_mock_cube(lat_dim_length=5, lon_dim_length=9))
        long_coord = gd.coord('longitude')
        long_coord.points = np.array([-180., -135., -90., -45., 0., 45., 90., 135., 180.])
        long_coord.bounds = None
        long_coord.guess_bounds()
        hpv = gd.get_non_masked_points()
        hpv.set_longitude_range(0.0)
        assert(np.min(hpv.longitudes) >= 0.0)
        assert(np.max(hpv.longitudes) < 360.0)
        assert(hpv[0, 0].longitude == 180.0)
        assert(hpv[4, 0].longitude == 180.0)
        assert(hpv[0, 1].longitude == 225.0)
        assert(hpv[4, 1].longitude == 225.0)
        assert(hpv[0, 4].longitude == 0.0)
        assert(hpv[4, 4].longitude == 0.0)
        assert(hpv[0, 8].longitude == 180.0)
        assert(hpv[4, 8].longitude == 180.0)


# if __name__ == '__main__':
#     import nose
#     nose.runmodule()



"""
module to test the NetCDF module
"""
import unittest
from netCDF4 import Variable

from hamcrest import *
from nose.tools import istest, raises, eq_

from cis.test.integration_test_data import *
from cis.data_io.netcdf import *

class TestNetCDF(unittest.TestCase):

    @raises(IOError)
    def should_raise_io_error_with_file_that_is_not_netcdf(self):
        read(valid_aeronet_filename, valid_aeronet_variable)

    def test_that_can_read_all_variables(self):
        dict = get_netcdf_file_variables(valid_2d_filename)
        eq_(len(dict), 426)

    def test_that_can_read_known_variable(self):
        data = read(valid_2d_filename, 'latitude')
        assert (isinstance(data['latitude'], Variable))

    def test_that_can_get_data(self):
        data = read(valid_2d_filename, 'latitude')
        eq_(data['latitude'].shape, (145,))

    def test_that_can_get_metadata_for_known_variable(self):
        data = read(valid_2d_filename, 'rain')
        metadata = get_metadata(data['rain'])

        eq_(str(metadata.missing_value), "2e+20")
        eq_(metadata.long_name, "TOTAL RAINFALL RATE: LS+CONV KG/M2/S")
        eq_(metadata.units,
            "kg m-2 s-1")

    def test_that_respect_valid_range_metadata(self):
        data = read(netcdf_cf_compliant_ranges_filename, netcdf_cf_compliant_ranges_vars)

        valid_min_data_array = get_data(data["CL_1944"])
        valid_min = data["CL_1944"].valid_min

        assert valid_min_data_array.min() >= valid_min

        valid_max_data_array = get_data(data["SO4_1950"])
        valid_max = data["SO4_1950"].valid_max

        assert valid_max_data_array.max() <= valid_max

        valid_range_data_array = get_data(data["LON_502"])
        valid_range = data["LON_502"].valid_range

        assert valid_range_data_array.min() >= valid_range[0]
        assert valid_range_data_array.max() <= valid_range[1]


class TestNetCDFGroups(unittest.TestCase):
    root_vars = ['StartOrbit', 'EndOrbit', 'SpaceCraftID', 'Year', 'Month', 'Day', 'Hour', 'Minute',
                 'DegradedInstMdr', 'DegradedProcMdr', 'GOMEScanMode']
    group_vars = {'GOME2': ['ChannelNumber',
                            'StartValidPixel',
                            'EndValidPixel',
                            'StartValidWavelengths',
                            'EndValidWavelengths',
                            'ChannelReadoutSeq',
                            'BandChannelNumber',
                            'BandNumber',
                            'StartPixel',
                            'NumberOfPixel',
                            'StartLambda',
                            'EndLambda',
                            'StartPixelPmd',
                            'LengthPixelPmd',
                            'WavelengthPmd'],
                  'AVHRR': ['Ch4CentralWavenumber',
                            'Ch4Constant1',
                            'Ch4Constant2Slope',
                            'Ch5CentralWavenumber',
                            'Ch5Constant1',
                            'iCh5Constant2Slope',
                            'ConstantC1',
                            'ConstantC2'],
                  'IASI': ['IASIFlag']}

    def _get_fully_qualified_vars(self):
        all_vars = self.root_vars
        for group, vars in self.group_vars.items():
            for var in vars:
                all_vars.append("/".join([group, var]))
        return all_vars

    def test_can_get_netcdf_file_variables(self):
        all_vars = self._get_fully_qualified_vars()
        nc_vars = get_netcdf_file_variables(valid_netcdf_groups_file)
        assert_that(set(nc_vars.keys()), is_(set(all_vars)))

    def test_can_read_many_files_individually(self):
        all_vars = self._get_fully_qualified_vars()
        data = read_many_files_individually([valid_netcdf_groups_file], all_vars)
        assert_that(data['StartOrbit'][0][:], is_([8031]))
        assert_that(data['GOME2/ChannelNumber'][0][:], contains(1, 2, 3, 4, 5, 6))
        assert_that(data['AVHRR/Ch4CentralWavenumber'][0][:], is_([933.63]))
        assert_that(data['IASI/IASIFlag'][0][:], is_([0]))

    def test_can_read(self):
        all_vars = self._get_fully_qualified_vars()
        data = read(valid_netcdf_groups_file, all_vars)
        assert_that(data['StartOrbit'][:], is_([8031]))
        assert_that(data['GOME2/ChannelNumber'][:], contains(1, 2, 3, 4, 5, 6))
        assert_that(data['AVHRR/Ch4CentralWavenumber'][:], is_([933.63]))
        assert_that(data['IASI/IASIFlag'][:], is_([0]))

    def test_can_remove_variables_with_non_spatiotemporal_dimensions(self):
        nc_vars = get_netcdf_file_variables(valid_netcdf_groups_file)
        allowed_dims = ['Dim4', 'Dim7', 'Dim8']
        remove_variables_with_non_spatiotemporal_dimensions(nc_vars, allowed_dims)
        expected_vars = ['DegradedInstMdr', 'DegradedProcMdr', 'GOMEScanMode',
                         'GOME2/StartPixelPmd', 'GOME2/LengthPixelPmd', 'GOME2/WavelengthPmd']
        assert_that(list(nc_vars.keys()), contains_inanyorder(*expected_vars))

    def test_can_read_nested_groups(self):
        var_name = 'group1/group2/var4'
        data = read(valid_nested_groups_file, var_name)
        assert_that(data[var_name][:], is_([12321]))

    def test_can_get_variables_nested_groups(self):
        nc_vars = get_netcdf_file_variables(valid_nested_groups_file)
        expected_vars = ['var1', 'var2', 'group1/var3', 'group1/group2/var4', 'group3/var5', 'group3/group4/var6']
        assert_that(list(nc_vars.keys()), contains_inanyorder(*expected_vars))



"""
Unit tests for the top-level subsetting routines.
"""
import datetime
from unittest import TestCase
import numpy as np
from cis.data_io.gridded_data import GriddedDataList

from cis.data_io.ungridded_data import UngriddedData, Metadata, UngriddedDataList
from cis.data_io.gridded_data import make_from_cube
import cis.test.util.mock


class TestGriddedSubsetConstraint(TestCase):
    """
    Tests for subsetting gridded data
    """

    def test_null_subset_2d_gridded_data(self):
        data = make_from_cube(cis.test.util.mock.make_square_5x3_2d_cube())
        subset = data.subset()
        np.testing.assert_array_equal(subset.data, data.data)

    def test_can_subset_2d_gridded_data_by_longitude(self):
        data = make_from_cube(cis.test.util.mock.make_square_5x3_2d_cube())
        subset = data.subset(longitude=[0.0, 5.0])
        assert (subset.data.tolist() == [[2, 3], [5, 6], [8, 9], [11, 12], [14, 15]])

    def test_can_subset_2d_gridded_data_by_latitude(self):
        data = make_from_cube(cis.test.util.mock.make_square_5x3_2d_cube())
        subset = data.subset(latitude=[0.0, 10.0])
        assert (subset.data.tolist() == [[7, 8, 9], [10, 11, 12], [13, 14, 15]])

    def test_can_subset_2d_gridded_data_by_latitude_with_None_min(self):
        data = make_from_cube(cis.test.util.mock.make_square_5x3_2d_cube())
        subset = data.subset(latitude=[0.0, None])
        assert (subset.data.tolist() == [[7, 8, 9], [10, 11, 12], [13, 14, 15]])

    def test_can_subset_2d_gridded_data_by_latitude_with_None_max(self):
        data = make_from_cube(cis.test.util.mock.make_square_5x3_2d_cube())
        subset = data.subset(latitude=[None, 0.0])
        assert (subset.data.tolist() == [[1, 2, 3], [4, 5, 6], [7, 8, 9]])

    def test_can_subset_3d_gridded_data_by_altitude(self):
        data = make_from_cube(cis.test.util.mock.make_square_5x3_2d_cube_with_altitude())
        subset = data.subset(altitude=[2, 5])
        assert (subset.data.tolist() == [[[3, 4, 5, 6], [10, 11, 12, 13], [17, 18, 19, 20]],
                                         [[24, 25, 26, 27], [31, 32, 33, 34], [38, 39, 40, 41]],
                                         [[45, 46, 47, 48], [52, 53, 54, 55], [59, 60, 61, 62]],
                                         [[66, 67, 68, 69], [73, 74, 75, 76], [80, 81, 82, 83]],
                                         [[87, 88, 89, 90], [94, 95, 96, 97], [101, 102, 103, 104]]])

    def test_can_subset_3d_gridded_data_by_pressure(self):
        data = make_from_cube(cis.test.util.mock.make_square_5x3_2d_cube_with_pressure())
        subset = data.subset(air_pressure=[2, 5])
        assert (subset.data.tolist() == [[[3, 4, 5, 6], [10, 11, 12, 13], [17, 18, 19, 20]],
                                         [[24, 25, 26, 27], [31, 32, 33, 34], [38, 39, 40, 41]],
                                         [[45, 46, 47, 48], [52, 53, 54, 55], [59, 60, 61, 62]],
                                         [[66, 67, 68, 69], [73, 74, 75, 76], [80, 81, 82, 83]],
                                         [[87, 88, 89, 90], [94, 95, 96, 97], [101, 102, 103, 104]]])

    def test_can_subset_2d_gridded_data_by_time(self):
        data = make_from_cube(cis.test.util.mock.make_square_5x3_2d_cube_with_time())
        subset = data.subset(time=[140494, 140497])
        assert (subset.data.tolist() == [[[3, 4, 5, 6], [10, 11, 12, 13], [17, 18, 19, 20]],
                                         [[24, 25, 26, 27], [31, 32, 33, 34], [38, 39, 40, 41]],
                                         [[45, 46, 47, 48], [52, 53, 54, 55], [59, 60, 61, 62]],
                                         [[66, 67, 68, 69], [73, 74, 75, 76], [80, 81, 82, 83]],
                                         [[87, 88, 89, 90], [94, 95, 96, 97], [101, 102, 103, 104]]])

    def test_can_subset_gridded_data_using_multiple_extract_constraints(self):
        from cis.time_util import convert_std_time_to_datetime
        data = make_from_cube(cis.test.util.mock.make_mock_cube(time_dim_length=3, alt_dim_length=6))
        subset = data.subset(time=convert_std_time_to_datetime([140492, 140493]), altitude=[0, 3])
        assert subset.data.shape == (5, 3, 3, 2)

    def test_can_subset_gridded_data_by_shape(self):
        data = make_from_cube(cis.test.util.mock.make_square_5x3_2d_cube())
        subset = data.subset(shape=cis.test.util.mock.WKT_DIAMOND)
        # The corner should be masked, but the center not
        assert np.ma.is_masked(subset.data[0, 0])
        assert ~np.ma.is_masked(subset.data[2, 1])
        assert (subset.data.tolist() == [[None, 5.0, None], 
                                         [7.0, 8.0, 9.0],
                                         [None, 11.0, None]])

    def test_can_subset_2d_gridded_data_by_longitude_with_wrapping_at_180(self):
        data = make_from_cube(cis.test.util.mock.make_mock_cube(lat_dim_length=5, lon_dim_length=9))
        long_coord = data.coord('longitude')
        long_coord.points = np.arange(-175, 185, 40)
        long_coord.bounds = None
        long_coord.guess_bounds()
        subset = data.subset(longitude=[135.0, 270])
        assert (subset.data.tolist() == [[9, 1, 2, 3],
                                         [18, 10, 11, 12],
                                         [27, 19, 20, 21],
                                         [36, 28, 29, 30],
                                         [45, 37, 38, 39]])

    def test_can_subset_2d_gridded_data_by_longitude_with_wrapping_at_360(self):
        data = make_from_cube(cis.test.util.mock.make_mock_cube(lat_dim_length=5, lon_dim_length=9))
        long_coord = data.coord('longitude')
        long_coord.points = np.arange(5, 365, 40)
        long_coord.bounds = None
        long_coord.guess_bounds()
        subset = data.subset(longitude=[-45.0, 90])
        assert (subset.data.tolist() == [[9, 1, 2, 3],
                                         [18, 10, 11, 12],
                                         [27, 19, 20, 21],
                                         [36, 28, 29, 30],
                                         [45, 37, 38, 39]])

    def test_can_subset_2d_gridded_data_with_missing_data(self):
        """This test just shows that missing values do not interfere with subsetting -
        nothing special happens to the missing values.
        """
        data = make_from_cube(cis.test.util.mock.make_square_5x3_2d_cube_with_missing_data())
        subset = data.subset(longitude=[0.0, 5.0])
        assert (subset.data.tolist(fill_value=-999) == [[2, 3], [-999, 6], [8, -999], [11, 12], [14, 15]])

    def test_can_subset_2d_gridded_data_by_longitude_latitude(self):
        data = make_from_cube(cis.test.util.mock.make_square_5x3_2d_cube())
        subset = data.subset(longitude=[0.0, 5.0], latitude=[-5.0, 5.0])
        assert (subset.data.tolist() == [[5, 6], [8, 9], [11, 12]])

    def test_edge_cases_with_no_bounds_for_2d_gridded_data_subset_by_longitude_latitude(self):
        """
        This test defines the behaviour for constraints set just away from cell centers when no bounds have been
        assigned to the coordinates: Namely the cell centers are treated as points with no bounds.

        This test constrains the subsetting algorithm as defined in the CIS paper so be very careful about changing it!
        """
        data = make_from_cube(cis.test.util.mock.make_square_5x3_2d_cube())
        subset = data.subset(longitude=[0.0, 4.5], latitude=[-5.5, 5.5])
        assert (subset.data.tolist() == [[5,], [8,], [11,]])

    def test_edge_cases_with_bounds_for_2d_gridded_data_subset_by_longitude_latitude(self):
        """
        Complementary to the test above, if bounds are assigned to the coordinates then the cells can match the
        constraint if the matching points are within - or equal to the bounds.

        This test constrains the subsetting algorithm as defined in the CIS paper so be very careful about changing it!
        """
        data = make_from_cube(cis.test.util.mock.make_square_5x3_2d_cube())
        for c in data.coords():
            c.guess_bounds()
        subset = data.subset(longitude=[0.0, 4.5], latitude=[-2.50, 2.50])
        assert (subset.data.tolist() == [[5, 6], [8, 9], [11, 12]])

    def test_empty_longitude_subset_of_gridded_data_list_returns_no_data(self):
        """
        Checks that the convention of returning None if subsetting results in an empty subset.
        Longitude has a modulus and so uses the IRIS intersection method
        """
        data = GriddedDataList([cis.test.util.mock.make_square_5x3_2d_cube()])
        subset = data.subset(longitude=[1.0, 3.0])
        assert (subset is None)

    def test_empty_time_subset_of_gridded_data_list_returns_no_data(self):
        """
        Checks that the convention of returning None if subsetting results in an empty subset.
        Longitude has no modulus and so uses the IRIS extract method
        """
        data = GriddedDataList([cis.test.util.mock.make_square_5x3_2d_cube_with_time()])
        subset = data.subset(time=[140500, 140550])
        assert (subset is None)

    def test_GIVEN_GriddedDataList_WHEN_constrain_THEN_correctly_subsetted_GriddedDataList_returned(self):
        gridded1 = cis.test.util.mock.make_square_5x3_2d_cube()
        gridded2 = cis.test.util.mock.make_square_5x3_2d_cube()
        datalist = GriddedDataList([gridded1, gridded2])
        subset = datalist.subset(longitude=[0.0, 5.0], latitude=[-5.0, 5.0])
        assert isinstance(subset, GriddedDataList)
        assert (subset[0].data.tolist() == [[5, 6], [8, 9], [11, 12]])
        assert (subset[1].data.tolist() == [[5, 6], [8, 9], [11, 12]])


class TestUngriddedSubsetConstraint(TestCase):
    # Tests for subsetting ungridded data

    def test_null_subset_of_2d_ungridded_data(self):
        data = cis.test.util.mock.make_regular_2d_ungridded_data()
        subset = data.subset()
        # The data gets flattened, but is otherwise equal
        np.testing.assert_array_equal(subset.data.flatten(), data.data.flatten())

    def test_can_subset_2d_ungridded_data_by_longitude(self):
        data = cis.test.util.mock.make_regular_2d_ungridded_data()
        subset = data.subset(longitude=[0.0, None])
        assert (subset.data.tolist() == [2, 3, 5, 6, 8, 9, 11, 12, 14, 15])

    def test_can_subset_2d_ungridded_data_by_longitude_with_None_min(self):
        data = cis.test.util.mock.make_regular_2d_ungridded_data()
        subset = data.subset(longitude=[None, 0.0])
        assert (subset.data.tolist() == [1, 2, 4, 5, 7, 8, 10, 11, 13, 14])

    def test_can_subset_2d_ungridded_data_by_longitude_with_wrapping_at_180(self):
        data = cis.test.util.mock.make_regular_2d_ungridded_data(
            lat_dim_length=5, lon_dim_length=9, lon_min=-175., lon_max=145.)
        subset = data.subset(longitude=[135.0, 270.0])
        assert (subset.data.tolist() == [1, 2, 3, 9, 10, 11, 12, 18, 19, 20, 21, 27, 28, 29, 30, 36, 37, 38, 39, 45])

    def test_can_subset_2d_ungridded_data_by_longitude_with_wrapping_at_360(self):
        data = cis.test.util.mock.make_regular_2d_ungridded_data(
            lat_dim_length=5, lon_dim_length=9, lon_min=5., lon_max=325.)
        subset = data.subset(longitude=[-45.0, 90.0])
        assert (subset.data.tolist() == [1, 2, 3, 9, 10, 11, 12, 18, 19, 20, 21, 27, 28, 29, 30, 36, 37, 38, 39, 45])

    def test_original_data_not_altered_when_subsetting(self):
        data = cis.test.util.mock.make_regular_2d_ungridded_data()
        subset = data.subset(longitude=[0.0, 5.0], latitude=[-5.0, 5.0])
        assert len(data.data_flattened) == 15
        assert len(data.coord('longitude').data_flattened) == 15

    def test_can_subset_2d_ungridded_data_by_longitude_latitude(self):
        data = cis.test.util.mock.make_regular_2d_ungridded_data()
        subset = data.subset(longitude=[0.0, 5.0], latitude=[-5.0, 5.0])
        assert (subset.data.tolist() == [5, 6, 8, 9, 11, 12])

    def test_empty_subset_of_ungridded_data_returns_no_data(self):
        """Checks that the convention of returning None if subsetting results in an empty subset.
        """
        data = cis.test.util.mock.make_regular_2d_ungridded_data()
        subset = data.subset(longitude=[1.0, 3.0])
        assert (subset is None)

    def test_can_subset_ungridded_data_by_time(self):
        data = cis.test.util.mock.make_regular_4d_ungridded_data()
        subset = data.subset(time=[datetime.datetime(1984, 8, 28), datetime.datetime(1984, 8, 29)])
        assert (subset.data.tolist() == [2, 3, 7, 8, 12, 13, 17, 18, 22, 23, 27, 28, 32, 33, 37, 38, 42, 43, 47, 48])

    def test_can_subset_ungridded_data_by_partial_date_time(self):
        from cis.time_util import PartialDateTime
        data = cis.test.util.mock.make_dummy_ungridded_data_time_series(100)
        subset = data.subset(time=[PartialDateTime(1984, 9)])
        assert (subset.data.tolist() == [6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0, 14.0, 15.0, 16.0, 17.0, 18.0, 19.0,
                                         20.0, 21.0, 22.0, 23.0, 24.0, 25.0, 26.0, 27.0, 28.0, 29.0, 30.0, 31.0, 32.0,
                                         33.0, 34.0, 35.0])

    def test_can_subset_ungridded_data_by_time_altitude(self):
        data = cis.test.util.mock.make_regular_4d_ungridded_data()
        subset = data.subset(time=[datetime.datetime(1984, 8, 28), datetime.datetime(1984, 8, 29)],
                             altitude=[45.0, 75.0])
        assert (subset.data.tolist() == [27, 28, 32, 33, 37, 38])

    def test_can_subset_ungridded_data_by_shape(self):
        data = cis.test.util.mock.make_regular_2d_with_time_ungridded_data()
        data.time.convert_to_std_time()
        subset = data.subset(shape=cis.test.util.mock.WKT_DIAMOND)
        # The corners should be masked, but the center not
        assert (subset.data.tolist() == [5, 7, 8, 9, 11])

    def test_can_subset_ungridded_data_by_shape_and_time(self):
        data = cis.test.util.mock.make_regular_2d_with_time_ungridded_data()
        data.time.convert_to_std_time()
        subset = data.subset(shape=cis.test.util.mock.WKT_DIAMOND, time=[datetime.datetime(1984, 8, 28),
                                                                         datetime.datetime(1984, 9, 3)])
        assert (subset.data.tolist() == [5, 7, 8])

    def test_can_subset_2d_ungridded_data_with_missing_values(self):
        data = cis.test.util.mock.make_regular_2d_ungridded_data_with_missing_values()
        subset = data.subset(longitude=[0.0, 5.0])
        assert (subset.data.tolist() == [2, 3, None, 6, 8, None, 11, 12, 14, 15])

    def test_GIVEN_UngriddedDataList_WHEN_constrain_THEN_correctly_subsetted_UngriddedDataList_returned(self):
        ug_data = cis.test.util.mock.make_regular_2d_ungridded_data()
        ug_data2 = UngriddedData(ug_data.data + 1, Metadata(name='snow', standard_name='snowfall_flux',
                                                            long_name="TOTAL SNOWFALL RATE: LS+CONV KG/M2/S",
                                                            units="kg m-2 s-1", missing_value=-999), ug_data.coords())
        datalist = UngriddedDataList([ug_data, ug_data2])
        subset = datalist.subset(longitude=[0.0, 5.0], latitude=[-5.0, 10.0])

        assert isinstance(subset, UngriddedDataList)
        assert subset[0].data.tolist() == [5, 6, 8, 9, 11, 12, 14, 15]
        assert subset[1].data.tolist() == [6, 7, 9, 10, 12, 13, 15, 16]
